#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-12 05:37:55 (ywatanabe)"
# _umap_dev.py


"""
This script does XYZ.
"""


"""
Imports
"""
import sys

import matplotlib.pyplot as plt
import mngs
import numpy as np
import umap.umap_ as umap_orig
from natsort import natsorted
from sklearn.preprocessing import LabelEncoder

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def umap(
    data,
    labels,
    hues=None,
    hues_colors=None,
    axes=None,
    axes_titles=None,
    supervised=False,
    title="UMAP Clustering",
    alpha=1.0,
    s=3,
    use_independent_legend=False,
    add_super_imposed=False,
    umap_model=None,
):
    """
    Perform UMAP clustering and visualization.

    Parameters
    ----------
    data_all : list
        List of data arrays to cluster
    labels_all : list
        List of label arrays corresponding to data_all
    hues_all : list, optional
        List of hue arrays for coloring points
    hues_colors_all : list, optional
        List of color mappings for hues
    axes : matplotlib.axes.Axes, optional
        Existing axes to plot on
    axes_titles : list, optional
        Titles for each subplot
    supervised : bool, optional
        Whether to use supervised UMAP
    title : str, optional
        Main title for the plot
    alpha : float, optional
        Transparency of points
    s : int, optional
        Size of points
    use_independent_legend : bool, optional
        Whether to create separate legend figures
    add_super_imposed : bool, optional
        Whether to add a superimposed plot
    umap_model : umap.UMAP, optional
        Pre-fitted UMAP model

    Returns
    -------
    tuple
        Figure, legend figures (if applicable), and UMAP model
    """

    # Renaming
    data_all = data
    labels_all = labels
    hues_all = hues
    hues_colors_all = hues_colors

    data_all, labels_all, hues_all, hues_colors_all = _check_input_vars(
        data_all, labels_all, hues_all, hues_colors_all
    )

    # Label Encoding
    le = LabelEncoder()
    le.fit(natsorted(np.hstack(labels_all)))
    labels_all = [le.transform(labels) for labels in labels_all]

    # Running UMAP Clustering
    _umap = _run_umap(umap_model, data_all, labels_all, supervised, title)

    # Plotting
    fig, legend_figs = _plot(
        _umap,
        le,
        data_all,
        labels_all,
        hues_all,
        hues_colors_all,
        add_super_imposed,
        axes,
        title,
        axes_titles,
        use_independent_legend,
        s,
        alpha,
    )

    return fig, legend_figs, _umap


def _plot(
    _umap,
    le,
    data_all,
    labels_all,
    hues_all,
    hues_colors_all,
    add_super_imposed,
    axes,
    title,
    axes_titles,
    use_independent_legend,
    s,
    alpha,
):
    # Plotting
    ncols = len(data_all) + 1 if add_super_imposed else len(data_all)
    share = True if ncols > 1 else False

    if axes is None:
        fig, axes = mngs.plt.subplots(ncols=ncols, sharex=share, sharey=share)
    else:
        assert len(axes) == ncols
        fig = (
            axes[0].get_figure()
            # axes
            if isinstance(
                axes, (np.ndarray, mngs.plt._subplots._AxesWrapper.AxesWrapper)
            )
            # axis
            else axes.get_figure()
        )

    fig.supxyt("UMAP 1", "UMAP 2", title)

    for ii, (data, labels, hues, hues_colors) in enumerate(
        zip(data_all, labels_all, hues_all, hues_colors_all)
    ):
        embedding = _umap.transform(data)

        # ax
        if ncols == 1:
            ax = axes
        else:
            ax = axes[ii + 1] if add_super_imposed else axes[ii]

        _hues = le.inverse_transform(labels) if hues is None else hues
        for hue in np.unique(_hues):
            indi = hue == np.array(_hues)

            if hues_colors:
                colors = np.vstack(hues_colors)[indi]
                colors = [colors[ii] for ii in range(len(colors))]
            else:
                colors = None
            ax.scatter(
                x=embedding[:, 0][indi],
                y=embedding[:, 1][indi],
                label=hue,
                c=colors,
                s=s,
                alpha=alpha,
            )

        ax.set_box_aspect(1)

        if axes_titles is not None:
            ax.set_title(axes_titles[ii])

        # Merged axis
        if add_super_imposed:
            ax = axes[0]
            _hues = le.inverse_transform(labels) if hues is None else hues
            for hue in np.unique(_hues):
                indi = hue == np.array(_hues)
                ax.scatter(
                    x=embedding[:, 0][indi],
                    y=embedding[:, 1][indi],
                    label=hue,
                    c=np.vstack(hues_colors)[indi][0],
                    s=s,
                    alpha=alpha,
                )

            ax.set_title("Superimposed")
            ax.set_box_aspect(1)
            # ax.sns_scatterplot(
            #     x=embedding[:, 0],
            #     y=embedding[:, 1],
            #     hue=le.inverse_transform(labels) if hues is None else hues,
            #     palette=hues_colors,
            #     legend="full" if ii == 0 else False,
            #     s=s,
            #     alpha=alpha,
            # )

    if share:
        mngs.plt.ax.sharex(axes)
        mngs.plt.ax.sharey(axes)

    if not use_independent_legend:
        for ax in axes.flat:
            ax.legend(loc="upper left")
        return fig, None

    elif use_independent_legend:
        legend_figs = []
        for i, ax in enumerate(axes):
            legend = ax.get_legend()
            if legend:
                legend_fig = plt.figure(figsize=(3, 2))

                new_legend = legend_fig.gca().legend(
                    handles=legend.get_lines(),
                    labels=[t.get_text() for t in legend.texts],
                    loc="center",
                )

                # new_legend = legend_fig.gca().legend(
                #     handles=legend.legendHandles,
                #     labels=legend.texts,
                #     loc="center",
                # )

                # legend_fig.canvas.draw()
                legend_figs.append(legend_fig)
                ax.get_legend().remove()

        for ax in axes:
            ax.legend_ = None

        # elif use_independent_legend:
        #     legend_figs = []
        #     for i, ax in enumerate(axes):
        #         legend = ax.get_legend()
        #         if legend:
        #             legend_fig = plt.figure(figsize=(3, 2))
        #             new_legend = legend_fig.gca().legend(
        #                 handles=legend.legendHandles,
        #                 labels=legend.texts,
        #                 loc="center",
        #             )
        #             legend_fig.canvas.draw()
        #             legend_filename = f"legend_{i}.png"
        #             legend_fig.savefig(legend_filename, bbox_inches="tight")
        #             legend_figs.append(legend_fig)
        #             plt.close(legend_fig)

        #     for ax in axes:
        #         ax.legend_ = None

    return fig, legend_figs


def _run_umap(umap_model, data_all, labels_all, supervised, title):
    # UMAP Clustering
    if not umap_model:
        umap_model = umap_orig.UMAP(random_state=42)
        supervised_label_or_none = labels_all[0] if supervised else None
        title = (
            f"(Supervised) {title}"
            if supervised
            else f"(Unsupervised) {title}"
        )
        _umap = umap_model.fit(data_all[0], y=supervised_label_or_none)
    else:
        _umap = umap_model

    return _umap


def _check_input_vars(data_all, labels_all, hues_all, hues_colors_all):
    # Ensures input formats
    if hues_all is None:
        hues_all = [None for _ in range(len(data_all))]

    if hues_colors_all is None:
        hues_colors_all = [None for _ in range(len(data_all))]

    assert (
        len(data_all)
        == len(labels_all)
        == len(hues_all)
        == len(hues_colors_all)
    )

    assert (
        isinstance(data_all, list)
        and isinstance(labels_all, list)
        and isinstance(hues_all, list)
        and isinstance(hues_colors_all, list)
    )
    return data_all, labels_all, hues_all, hues_colors_all


def _test(dataset_str="iris"):
    import matplotlib.pyplot as plt
    import numpy as np
    from sklearn.datasets import load_digits, load_iris
    from sklearn.model_selection import train_test_split

    # Load iris dataset
    load_dataset = {"iris": load_iris, "mnist": load_digits}[dataset_str]

    dataset = load_dataset()
    X = dataset.data
    y = dataset.target

    # Split data into two parts
    X1, X2, y1, y2 = train_test_split(X, y, test_size=0.5, random_state=42)

    # Call umap function
    fig, legend_figs, umap_model = umap(
        data=[X1, X2],
        labels=[y1, y2],
        # axes=axes,
        axes_titles=[f"{dataset_str} Set 1", f"{dataset_str} Set 2"],
        supervised=True,
        title=dataset_str,
        use_independent_legend=True,
        s=10,
    )

    # plt.tight_layout()
    mngs.io.save(fig, f"/tmp/mngs/umap/{dataset_str}.jpg")

    # Save legend figures if any
    if legend_figs:
        for i, leg_fig in enumerate(legend_figs):
            mngs.io.save(
                leg_fig, f"/tmp/mngs/umap/{dataset_str}_legend_{i}.jpg"
            )


main = umap

if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False, agg=True
    )
    _test(dataset_str="mnist")
    # main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-30 07:27:27 (ywatanabe)"

import torch
import torch.nn as nn
import torch.nn.functional as F


class AxiswiseDropout(nn.Module):
    def __init__(self, dropout_prob=0.5, dim=1):
        super(AxiswiseDropout, self).__init__()
        self.dropout_prob = dropout_prob
        self.dim = dim

    def forward(self, x):
        if self.training:
            sizes = [s if i == self.dim else 1 for i, s in enumerate(x.size())]
            dropout_mask = F.dropout(
                torch.ones(*sizes, device=x.device, dtype=x.dtype),
                self.dropout_prob,
                True,
            )

            # Expand the mask to the size of the input tensor and apply it
            return x * dropout_mask.expand_as(x)
        return x



#!/usr/bin/env python3

import fnmatch
import glob
import inspect
import os
import re
import subprocess
import warnings
from glob import glob

import mngs


################################################################################
## PATH
################################################################################
def get_this_fpath(when_ipython="/tmp/fake.py"):
    """
    Get the file path of the calling script, with special handling for IPython environments.

    Parameters:
    -----------
    when_ipython : str, optional
        The file path to return when running in an IPython environment. Default is "/tmp/fake.py".

    Returns:
    --------
    str
        The file path of the calling script or the specified path for IPython environments.

    Example:
    --------
    >>> import mngs.io._path as path
    >>> fpath = path.get_this_fpath()
    >>> print(fpath)
    '/path/to/current/script.py'
    """
    __file__ = inspect.stack()[1].filename
    if "ipython" in __file__:  # for ipython
        __file__ = when_ipython  # "/tmp/fake.py"
    return __file__


def mk_spath(sfname, makedirs=False):
    """
    Create a save path based on the calling script's location.

    Parameters:
    -----------
    sfname : str
        The name of the file to be saved.
    makedirs : bool, optional
        If True, create the directory structure for the save path. Default is False.

    Returns:
    --------
    str
        The full save path for the file.

    Example:
    --------
    >>> import mngs.io._path as path
    >>> spath = path.mk_spath('output.txt', makedirs=True)
    >>> print(spath)
    '/path/to/current/script/output.txt'
    """
    __file__ = inspect.stack()[1].filename
    if "ipython" in __file__:  # for ipython
        __file__ = f'/tmp/fake-{os.getenv("USER")}.py'

    ## spath
    fpath = __file__
    fdir, fname, _ = split_fpath(fpath)
    sdir = fdir + fname + "/"
    spath = sdir + sfname

    if makedirs:
        os.makedirs(mngs.path.split(spath)[0], exist_ok=True)

    return spath


def find_the_git_root_dir():
    """
    Find the root directory of the current Git repository.

    Returns:
    --------
    str
        The path to the root directory of the current Git repository.

    Raises:
    -------
    git.exc.InvalidGitRepositoryError
        If the current directory is not part of a Git repository.

    Example:
    --------
    >>> import mngs.io._path as path
    >>> git_root = path.find_the_git_root_dir()
    >>> print(git_root)
    '/path/to/git/repository'
    """
    import git

    repo = git.Repo(".", search_parent_directories=True)
    return repo.working_tree_dir


def split_fpath(fpath):
    """
    Split a file path into directory path, file name, and file extension.

    Parameters:
    -----------
    fpath : str
        The full file path to split.

    Returns:
    --------
    tuple
        A tuple containing (dirname, fname, ext), where:
        - dirname: str, the directory path
        - fname: str, the file name without extension
        - ext: str, the file extension

    Example:
    --------
    >>> dirname, fname, ext = split_fpath('../data/01/day1/split_octave/2kHz_mat/tt8-2.mat')
    >>> print(dirname)
    '../data/01/day1/split_octave/2kHz_mat/'
    >>> print(fname)
    'tt8-2'
    >>> print(ext)
    '.mat'
    """
    dirname = os.path.dirname(fpath) + "/"
    base = os.path.basename(fpath)
    fname, ext = os.path.splitext(base)
    return dirname, fname, ext


def touch(fpath):
    """
    Create a file or update its modification time.

    This function mimics the Unix 'touch' command.

    Parameters:
    -----------
    fpath : str
        The path to the file to be touched.

    Returns:
    --------
    None

    Side Effects:
    -------------
    Creates a new file if it doesn't exist, or updates the modification time
    of an existing file.

    Example:
    --------
    >>> import mngs.io._path as path
    >>> import os
    >>> test_file = '/tmp/test_file.txt'
    >>> path.touch(test_file)
    >>> assert os.path.exists(test_file)
    >>> print(f"File created: {test_file}")
    File created: /tmp/test_file.txt
    """
    import pathlib

    return pathlib.Path(fpath).touch()


def find(rootdir, type="f", exp=["*"]):
    """
    Mimic the Unix find command to search for files or directories.

    Parameters:
    -----------
    rootdir : str
        The root directory to start the search from.
    type : str, optional
        The type of entries to search for. 'f' for files, 'd' for directories,
        or None for both. Default is 'f'.
    exp : str or list of str, optional
        Pattern(s) to match against file or directory names. Default is ["*"].

    Returns:
    --------
    list
        A list of paths that match the search criteria.

    Example:
    --------
    >>> find('/path/to/search', "f", "*.txt")
    ['/path/to/search/file1.txt', '/path/to/search/subdir/file2.txt']
    """
    if isinstance(exp, str):
        exp = [exp]

    matches = []
    for _exp in exp:
        for root, dirs, files in os.walk(rootdir):
            # Depending on the type, choose the list to iterate over
            if type == "f":  # Files only
                names = files
            elif type == "d":  # Directories only
                names = dirs
            else:  # All entries
                names = files + dirs

            for name in names:
                # Construct the full path
                path = os.path.join(root, name)

                # If an _exp is provided, use fnmatch to filter names
                if _exp and not fnmatch.fnmatch(name, _exp):
                    continue

                # If type is set, ensure the type matches
                if type == "f" and not os.path.isfile(path):
                    continue
                if type == "d" and not os.path.isdir(path):
                    continue

                # Add the matching path to the results
                matches.append(path)

    return matches


def find_latest(dirname, fname, ext, version_prefix="_v"):
    """
    Find the latest version of a file with a specific naming pattern.

    This function searches for files in the given directory that match the pattern:
    {fname}{version_prefix}{number}{ext} and returns the path of the file with the highest version number.

    Parameters:
    -----------
    dirname : str
        The directory to search in.
    fname : str
        The base filename without version number or extension.
    ext : str
        The file extension, including the dot (e.g., '.txt').
    version_prefix : str, optional
        The prefix used before the version number. Default is '_v'.

    Returns:
    --------
    str or None
        The full path of the latest version file if found, None otherwise.

    Example:
    --------
    >>> find_latest('/path/to/dir', 'myfile', '.txt')
    '/path/to/dir/myfile_v3.txt'
    """
    version_pattern = re.compile(
        rf"({re.escape(fname)}{re.escape(version_prefix)})(\d+)({re.escape(ext)})$"
    )

    glob_pattern = os.path.join(dirname, f"{fname}{version_prefix}*{ext}")
    files = glob(glob_pattern)

    highest_version = 0
    latest_file = None

    for file in files:
        filename = os.path.basename(file)
        match = version_pattern.search(filename)
        if match:
            version_num = int(match.group(2))
            if version_num > highest_version:
                highest_version = version_num
                latest_file = file

    return latest_file


def increment_version(dirname, fname, ext, version_prefix="_v"):
    """
    Generate the next version of a filename based on existing versioned files.

    This function searches for files in the given directory that match the pattern:
    {fname}{version_prefix}{number}{ext} and returns the path for the next version.

    Parameters:
    -----------
    dirname : str
        The directory to search in and where the new file will be created.
    fname : str
        The base filename without version number or extension.
    ext : str
        The file extension, including the dot (e.g., '.txt').
    version_prefix : str, optional
        The prefix used before the version number. Default is '_v'.

    Returns:
    --------
    str
        The full path for the next version of the file.

    Example:
    --------
    >>> increment_version('/path/to/dir', 'myfile', '.txt')
    '/path/to/dir/myfile_v004.txt'

    Notes:
    ------
    - If no existing versioned files are found, it starts with version 001.
    - The version number is always formatted with at least 3 digits.
    """
    # Create a regex pattern to match the version number in the filename
    version_pattern = re.compile(
        rf"({re.escape(fname)}{re.escape(version_prefix)})(\d+)({re.escape(ext)})$"
    )

    # Construct the glob pattern to find all files that match the pattern
    glob_pattern = os.path.join(dirname, f"{fname}{version_prefix}*{ext}")

    # Use glob to find all files that match the pattern
    files = glob(glob_pattern)

    # Initialize the highest version number
    highest_version = 0
    base, suffix = None, None

    # Loop through the files to find the highest version number
    for file in files:
        filename = os.path.basename(file)
        match = version_pattern.search(filename)
        if match:
            base, version_str, suffix = match.groups()
            version_num = int(version_str)
            if version_num > highest_version:
                highest_version = version_num

    # If no versioned files were found, use the provided filename and extension
    if base is None or suffix is None:
        base = f"{fname}{version_prefix}"
        suffix = ext
        highest_version = 0  # No previous versions

    # Increment the highest version number
    next_version_number = highest_version + 1

    # Format the next version number with the same number of digits as the original
    next_version_str = f"{base}{next_version_number:03d}{suffix}"

    # Combine the directory and new filename to create the full path
    next_filepath = os.path.join(dirname, next_version_str)

    return next_filepath



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-10 23:11:58 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/_gen_ai/_format_output_func.py


"""
This script does XYZ.
"""


"""
Imports
"""
import re
import sys

import markdown2
import matplotlib.pyplot as plt
import mngs

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def format_output_func(out_text):
    def find_unwrapped_urls(text):
        # Regex to find URLs that are not already within <a href> tags
        url_pattern = (
            r'(?<!<a href=")(https?://|doi:|http://doi.org/)[^\s,<>"]+'
        )

        # Find all matches that are not already wrapped
        unwrapped_urls = re.findall(url_pattern, text)

        return unwrapped_urls

    def add_a_href_tag(text):
        # Function to replace each URL with its wrapped version
        def replace_url(match):
            url = match.group(0)
            # Normalize DOI URLs
            if url.startswith("doi:"):
                url = "https://doi.org/" + url[4:]
            return f'<a href="{url}">{url}</a>'

        # Regex pattern to match URLs not already wrapped in <a> tags
        url_pattern = (
            r'(?<!<a href=")(https?://|doi:|http://doi.org/)[^\s,<>"]+'
        )

        # Replace all occurrences of unwrapped URLs in the text
        updated_text = re.sub(url_pattern, replace_url, text)

        return updated_text

    def add_masked_api_key(text, api_key):
        masked_api_key = f"{api_key[:4]}****{api_key[-4:]}"
        return text + f"\n(API Key: {masked_api_key}"

    out_text = markdown2.markdown(out_text)
    out_text = add_a_href_tag(out_text)
    out_text = re.sub(r"^<p>(.*)</p>$", r"\1", out_text, flags=re.DOTALL)
    return out_text


def main():
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-05 18:56:18 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/plt/_subplots/_to_sigma.py


"""
This script does XYZ.
"""


"""
Imports
"""
import os
import re
import sys

import matplotlib
import matplotlib.pyplot as plt
import mngs
import seaborn as sns

mngs.gen.reload(mngs)
import logging
import warnings
from glob import glob
from pprint import pprint

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import xarray as xr
from icecream import ic
from natsort import natsorted
from tqdm import tqdm

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""
import mngs
import pandas as pd


def to_sigma(history):
    if len(history) > 0:
        values_list = list(history.values())
        data_frames = [format_plotting_args(record) for record in values_list]
        try:
            combined_data = pd.concat(data_frames, axis=1)
            return combined_data
        except Exception as e:
            print(e)
            return pd.DataFrame()
    else:
        return


def format_plotting_args(record):
    id, method, args, kwargs = record

    # Matplotlib
    if method in ["plot"]:
        if len(args) == 1:
            args = args[0]
            if args.ndim == 2:
                x, y = args[:, 0], args[:, 1]
                df = pd.DataFrame({f"{id}_{method}_x": x})
                return df

        elif len(args) == 2:
            x, y = args

            if isinstance(y, (np.ndarray, xr.DataArray)):
                if y.ndim == 2:
                    from collections import OrderedDict

                    out = OrderedDict()

                    for ii in range(y.shape[1]):
                        out[f"{id}_{method}_x{ii:02d}"] = x
                        out[f"{id}_{method}_y{ii:02d}"] = y[:, ii]
                    df = pd.DataFrame(out)

                    return df

            if isinstance(y, pd.DataFrame):
                df = pd.DataFrame(
                    {
                        f"{id}_{method}_x": x,
                        **{
                            f"{id}_{method}_y{ii:02d}": np.array(y[col])
                            for ii, col in enumerate(y.columns)
                        },
                    }
                )
                return df

            else:
                if isinstance(y, (np.ndarray, xr.DataArray, list)):
                    df = pd.DataFrame(
                        {f"{id}_{method}_x": x, f"{id}_{method}_y": y}
                    )
                    return df

    elif method == "scatter":
        x, y = args
        df = pd.DataFrame({f"{id}_{method}_x": x, f"{id}_{method}_y": y})
        return df

    elif method == "bar":
        x, y = args
        yerr = kwargs.get("yerr")

        if isinstance(x, (int, float)):
            x = pd.Series(x, name="x")
        if isinstance(y, (int, float)):
            y = pd.Series(y, name="y")

        df = pd.DataFrame({f"{id}_{method}_x": x, f"{id}_{method}_y": y})

        if yerr is not None:
            if isinstance(yerr, (int, float)):
                yerr = pd.Series(yerr, name="yerr")
            df[f"{id}_{method}_yerr"] = yerr
        return df

    elif method == "hist":
        x = args
        df = pd.DataFrame({f"{id}_{method}_x": x})
        return df

    elif method == "boxplot":
        x = args[0]

        # One box plot
        if isinstance(x, np.ndarray) or mngs.gen.is_listed_X(x, [float, int]):
            df = pd.DataFrame(x)

        else:
            # Multiple boxes
            df = mngs.pd.force_df({i_x: _x for i_x, _x in enumerate(x)})
        df.columns = [f"{id}_{method}_{col}_x" for col in df.columns]
        df = df.apply(lambda col: col.dropna().reset_index(drop=True))
        return df

    # elif method == "boxplot_":
    #     __import__("ipdb").set_trace()
    #     x = args[0]
    #     df =
    #     df.columns = [f"{id}_{method}_{col}" for col in df.columns]

    #     return df

    elif method == "plot_":
        df = args
        df.columns = [f"{id}_{method}_{col}" for col in df.columns]
        return df

    elif method == "fillv":
        starts, ends = args
        df = pd.DataFrame(
            {
                f"{id}_{method}_starts": starts,
                f"{id}_{method}_ends": ends,
            }
        )
        return df

    elif method == "raster":
        df = args
        return df

    elif method == "ecdf":
        df = args
        return df

    elif method == "kde":
        df = args
        if id is not None:
            df.columns = [f"{id}_{method}_{col}" for col in df.columns]
        return df

    elif method == "imshow2d":
        df = args
        return df

    elif method == "sns_barplot":
        __import__("ipdb").set_trace()
        return df

    elif method == "sns_boxplot":
        df = args
        if id is not None:
            df.columns = [f"{id}_{method}_{col}" for col in df.columns]
        return df

    elif method == "sns_heatmap":
        df = args
        return df

    elif method == "sns_histplot":
        df = args
        return df

    elif method == "sns_kdeplot":
        pass
        # df = args
        # __import__("ipdb").set_trace()
        # return df

    elif method == "sns_lineplot":
        __import__("ipdb").set_trace()
        return df

    elif method == "sns_pairplot":
        __import__("ipdb").set_trace()
        return df

    elif method == "sns_scatterplot":
        __import__("ipdb").set_trace()
        return df

    elif method == "sns_violinplot":
        df = args
        return df

    elif method == "sns_jointplot":
        __import__("ipdb").set_trace()
        return df

    else:
        pass
        # if not method.startswith("set_"):
        #     logging.warn(
        #         f"{method} is not implemented in _to_sigma method of the mngs.plt module."
        #     )


def main():
    # Line
    fig, ax = mngs.plt.subplots()
    ax.plot([1, 2, 3], [4, 5, 6], id="plot1")
    ax.plot([4, 5, 6], [1, 2, 3], id="plot2")
    mngs.io.save(fig, "./plots.png")
    mngs.io.save(ax.to_sigma(), "./plots.csv")

    # No tracking
    fig, ax = mngs.plt.subplots(track=False)
    ax.plot([1, 2, 3], [4, 5, 6], id="plot3")
    ax.plot([4, 5, 6], [1, 2, 3], id="plot4")
    mngs.io.save(fig, "./plots_wo_tracking.png")
    mngs.io.save(ax.to_sigma(), "./plots_wo_tracking.csv")

    # Scatter
    fig, ax = mngs.plt.subplots()
    ax.scatter([1, 2, 3], [4, 5, 6], id="scatter1")
    ax.scatter([4, 5, 6], [1, 2, 3], id="scatter2")
    mngs.io.save(fig, "./scatters.png")
    mngs.io.save(ax.to_sigma(), "./scatters.csv")

    # Box
    fig, ax = mngs.plt.subplots()
    ax.boxplot([1, 2, 3], id="boxplot1")
    mngs.io.save(fig, "./boxplot1.png")
    mngs.io.save(ax.to_sigma(), "./boxplot1.csv")

    # Bar
    fig, ax = mngs.plt.subplots()
    ax.bar(["A", "B", "C"], [4, 5, 6], id="bar1")
    mngs.io.save(fig, "./bar1.png")
    mngs.io.save(ax.to_sigma(), "./bar1.csv")

    # # Bar
    # # fig, ax = mngs.plt.subplots()
    # fig, ax = plt.subplots()
    # ax.bar(["A", "B", "C"], [4, 5, 6], id="bar2")
    # mngs.io.save(fig, "./bar2.png")
    # mngs.io.save(ax.to_sigma(), "./bar2.csv")

    # print(ax.to_sigma())
    # #    plot1_plot_x  plot1_plot_y  plot2_plot_x  ...  boxplot1_boxplot_x  bar1_bar_x  bar1_bar_y
    # # 0           1.0           4.0           4.0  ...                 1.0           A         4.0
    # # 1           2.0           5.0           5.0  ...                 2.0           B         5.0
    # # 2           3.0           6.0           6.0  ...                 3.0           C         6.0

    # print(ax.to_sigma().keys())  # plot3 and plot 4 are not tracked
    # # [3 rows x 11 columns]
    # # Index(['plot1_plot_x', 'plot1_plot_y', 'plot2_plot_x', 'plot2_plot_y',
    # #        'scatter1_scatter_x', 'scatter1_scatter_y', 'scatter2_scatter_x',
    # #        'scatter2_scatter_y', 'boxplot1_boxplot_x', 'bar1_bar_x', 'bar1_bar_y'],
    # #       dtype='object')

    # # If a path is passed, the sigmaplot-friendly dataframe is saved as a csv file.
    # ax.to_sigma("../for_sigmaplot.csv")
    # # Saved to: ../for_sigmaplot.csv


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False, agg=True
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-06 10:39:57 (ywatanabe)"

import pandas as pd
import re
from typing import Union, List

def p2stars(input_data: Union[float, str, pd.DataFrame], ns: bool = False) -> Union[str, pd.DataFrame]:
    """
    Convert p-value(s) to significance stars.

    Example
    -------
    >>> p2stars(0.0005)
    '***'
    >>> p2stars("0.03")
    '*'
    >>> p2stars("1e-4")
    '***'
    >>> df = pd.DataFrame({'p_value': [0.001, "0.03", 0.1, "NA"]})
    >>> p2stars(df)
       p_value
    0    0.001 ***
    1    0.030   *
    2    0.100
    3       NA  NA

    Parameters
    ----------
    input_data : float, str, or pd.DataFrame
        The p-value or DataFrame containing p-values to convert.
        For DataFrame, columns matching re.search(r'p[_.-]?val', col.lower()) are considered.
    ns : bool, optional
        Whether to return 'n.s.' for non-significant results (default is False)

    Returns
    -------
    str or pd.DataFrame
        Significance stars or DataFrame with added stars column
    """
    if isinstance(input_data, (float, int, str)):
        return _p2stars_str(input_data, ns)
    elif isinstance(input_data, pd.DataFrame):
        return _p2stars_pd(input_data, ns)
    else:
        raise ValueError("Input must be a float, string, or a pandas DataFrame")

def _p2stars_str(pvalue: Union[float, str], ns: bool = False) -> str:
    try:
        if isinstance(pvalue, str):
            pvalue = pvalue.strip().lower()
            if pvalue in ['na', 'nan', 'null', '']:
                return 'NA'
        pvalue_float = float(pvalue)
        if pvalue_float < 0 or pvalue_float > 1:
            raise ValueError(f"P-value must be between 0 and 1, got {pvalue_float}")
    except ValueError as e:
        raise ValueError(f"Invalid p-value: {pvalue}. {str(e)}")

    if pvalue_float <= 0.001:
        return "***"
    elif pvalue_float <= 0.01:
        return "**"
    elif pvalue_float <= 0.05:
        return "*"
    else:
        return "n.s." if ns else ""


def _p2stars_pd(df: pd.DataFrame, ns: bool = False) -> pd.DataFrame:
    from mngs.stats import find_pval
    pvalue_cols = find_pval(df, multiple=True)
    assert pvalue_cols, "No p-value columns found in DataFrame"

    for pvalue_col in pvalue_cols:
        star_col = pvalue_col + "_stars"
        df[star_col] = df[pvalue_col].apply(lambda x: _p2stars_str(x, ns))

        # Get the index of the current p-value column
        col_idx = df.columns.get_loc(pvalue_col)

        # Move the star column right after the p-value column
        cols = list(df.columns)
        cols.insert(col_idx + 1, cols.pop(cols.index(star_col)))
        df = df.reindex(columns=cols)

    return df

# def _find_pvalue_columns(df: pd.DataFrame) -> List[str]:
#     """
#     Find columns that likely contain p-values.

#     Example
#     -------
#     >>> df = pd.DataFrame({'p_value': [0.05], 'pval': [0.01], 'p-val': [0.001], 'p.value': [0.1]})
#     >>> _find_pvalue_columns(df)
#     ['p_value', 'pval', 'p-val', 'p.value']

#     Parameters
#     ----------
#     df : pd.DataFrame
#         Input DataFrame

#     Returns
#     -------
#     List[str]
#         List of column names that likely contain p-values
#     """
#     return [col for col in df.columns if re.search(r'p[_.-]?val', col.lower())]



import sys
import os

def flush(sys=sys):
    """
    Flushes the system's stdout and stderr, and syncs the file system.
    This ensures all pending write operations are completed.
    """
    sys.stdout.flush()
    sys.stderr.flush()
    os.sync()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-14 00:28:03 (ywatanabe)"
# File: 2024-10-07 22:17:30

try:
    from ._typing import List, Tuple, Dict, Any, Union, Sequence, Literal, Optional, Iterable, Generator, ArrayLike
except ImportError as e:
    warnings.warn(f"Warning: Failed to import from ._typing.")

# from ._typing import List, Tuple, Dict, Any, Union, Sequence, Literal, Optional, Iterable, Generator, ArrayLike



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-02-11 21:04:35 (ywatanabe)"

import pandas as pd
import numpy as np

def add_hue(df):
    df["hue"] = 0
    dummy_row = pd.DataFrame(
        columns=df.columns,
        data=np.array([np.nan for _ in df.columns]).reshape(1, -1),
    )
    dummy_row = {}
    for col in df.columns:
        dtype = df[col].dtype
        if dtype is np.dtype(object):
            dummy_row[col] = np.nan
        if dtype is np.dtype(float):
            dummy_row[col] = np.nan
        if dtype is np.dtype(np.int64):
            dummy_row[col] = np.nan
        if dtype is np.dtype(bool):
            dummy_row[col] = None

    dummy_row = pd.DataFrame(pd.Series(dummy_row)).T

    dummy_row["hue"] = 1
    df_added = pd.concat([df, dummy_row], axis=0)
    return df_added



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-03 06:04:45 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/pd/_mv.py


def mv(df, key, position, axis=1):
    """
    Move a row or column to a specified position in a DataFrame.

    Args:
    df (pandas.DataFrame): The input DataFrame.
    key (str): The label of the row or column to move.
    position (int): The position to move the row or column to.
    axis (int, optional): 0 for rows, 1 for columns. Defaults to 1.

    Returns:
    pandas.DataFrame: A new DataFrame with the row or column moved.
    """
    if axis == 0:
        items = df.index.tolist()
    else:
        items = df.columns.tolist()
    items.remove(key)

    if position < 0:
        position += len(items) + 1

    items.insert(position, key)
    return df.reindex(items, axis=axis)


def mv_to_first(df, key, axis=0):
    """
    Move a row or column to the first position in a DataFrame.

    Args:
    df (pandas.DataFrame): The input DataFrame.
    key (str): The label of the row or column to move.
    axis (int, optional): 0 for rows, 1 for columns. Defaults to 0.

    Returns:
    pandas.DataFrame: A new DataFrame with the row or column moved to the first position.
    """
    return mv(df, key, 0, axis)


def mv_to_last(df, key, axis=0):
    """
    Move a row or column to the last position in a DataFrame.

    Args:
    df (pandas.DataFrame): The input DataFrame.
    key (str): The label of the row or column to move.
    axis (int, optional): 0 for rows, 1 for columns. Defaults to 0.

    Returns:
    pandas.DataFrame: A new DataFrame with the row or column moved to the last position.
    """
    return mv(df, key, -1, axis)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-06 19:04:03 (ywatanabe)"#!/usr/bin/env python3

"""
Functionality:
    - Provides utility functions and decorators for seamless data type conversions between NumPy, PyTorch, and Pandas
    - Implements batch processing functionality for large datasets
Input:
    - Various data types including NumPy arrays, PyTorch tensors, Pandas DataFrames, and Python lists
Output:
    - Converted data in the desired format (NumPy, PyTorch, or Pandas)
Prerequisites:
    - NumPy, PyTorch, Pandas, xarray, tqdm
"""

import warnings
from functools import wraps, lru_cache
from typing import Union, Tuple, List, Dict, Any, Optional, Callable

import numpy as np
import pandas as pd
import torch
import xarray
from tqdm import tqdm

ArrayLike = Union[List, Tuple, np.ndarray, pd.Series, pd.DataFrame, xarray.DataArray, torch.Tensor]

class ConversionWarning(UserWarning):
    pass

@lru_cache(maxsize=None)
def _cached_warning(message: str) -> None:
    warnings.warn(message, category=ConversionWarning)

def _conversion_warning(old: Any, new: torch.Tensor) -> None:
    message = (
        f"Converted from {type(old).__name__} to {type(new).__name__} ({new.device}). "
        f"Consider using {type(new).__name__} ({new.device}) as input for faster computation."
    )
    _cached_warning(message)

warnings.simplefilter("always", ConversionWarning)

def try_device(tensor: torch.Tensor, device: str) -> torch.Tensor:
    if not isinstance(tensor, torch.Tensor):
        return tensor
    if tensor.device.type == device:
        return tensor
    try:
        return tensor.to(device)
    except RuntimeError as error:
        if "cuda" in str(error).lower() and device == "cuda":
            warnings.warn("CUDA memory insufficient, falling back to CPU.", UserWarning)
            return tensor.cpu()
        raise error

def unsqueeze_if(arr_or_tensor: Union[np.ndarray, torch.Tensor], ndim: int = 2, axis: int = 0) -> torch.Tensor:
    if not isinstance(arr_or_tensor, (np.ndarray, torch.Tensor)):
        raise TypeError("Input must be a NumPy array or a PyTorch tensor.")
    if arr_or_tensor.ndim != ndim:
        raise ValueError(f"Input must be a {ndim}-D array or tensor.")
    return torch.unsqueeze(arr_or_tensor, dim=axis)

def squeeze_if(arr_or_tensor: Union[np.ndarray, torch.Tensor], ndim: int = 3, axis: int = 0) -> torch.Tensor:
    if not isinstance(arr_or_tensor, (np.ndarray, torch.Tensor)):
        raise TypeError("Input must be a NumPy array or a PyTorch tensor.")
    if arr_or_tensor.ndim != ndim:
        raise ValueError(f"Input must be a {ndim}-D array or tensor.")
    return torch.squeeze(arr_or_tensor, dim=axis)

def is_torch(*args: Any, **kwargs: Any) -> bool:
    return any(isinstance(arg, torch.Tensor) for arg in args) or any(isinstance(val, torch.Tensor) for val in kwargs.values())

def is_cuda(*args: Any, **kwargs: Any) -> bool:
    return any((isinstance(arg, torch.Tensor) and arg.is_cuda) for arg in args) or any((isinstance(val, torch.Tensor) and val.is_cuda) for val in kwargs.values())

def return_always(*args: Any, **kwargs: Any) -> Tuple[Tuple, Dict]:
    return args, kwargs

def return_if(*args: Any, **kwargs: Any) -> Union[Tuple, Dict, None]:
    if args and kwargs:
        return args, kwargs
    elif args:
        return args
    elif kwargs:
        return kwargs
    else:
        return None

def to_torch(*args: Any, return_fn: Callable = return_if, **kwargs: Any) -> Any:
    def _to_torch(data: Any, device: Optional[str] = kwargs.get("device")) -> Any:
        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"

        if isinstance(data, (pd.Series, pd.DataFrame)):
            new_data = torch.tensor(data.to_numpy()).squeeze().float()
            new_data = try_device(new_data, device)
            if device == "cuda":
                _conversion_warning(data, new_data)
            return new_data

        if isinstance(data, (np.ndarray, list)):
            new_data = torch.tensor(data).float()
            new_data = try_device(new_data, device)
            if device == "cuda":
                _conversion_warning(data, new_data)
            return new_data

        if isinstance(data, xarray.core.dataarray.DataArray):
            new_data = torch.tensor(np.array(data)).float()
            new_data = try_device(new_data, device)
            if device == "cuda":
                _conversion_warning(data, new_data)
            return new_data

        if isinstance(data, tuple):
            return [_to_torch(item) for item in data if item is not None]

        return data

    converted_args = [_to_torch(arg) for arg in args if arg is not None]
    converted_kwargs = {key: _to_torch(val) for key, val in kwargs.items() if val is not None}

    if "axis" in converted_kwargs:
        converted_kwargs["dim"] = converted_kwargs.pop("axis")

    return return_fn(*converted_args, **converted_kwargs)

def to_numpy(*args: Any, return_fn: Callable = return_if, **kwargs: Any) -> Any:
    def _to_numpy(data: Any) -> Any:
        if isinstance(data, (pd.Series, pd.DataFrame)):
            return data.to_numpy().squeeze()
        if isinstance(data, torch.Tensor):
            return data.detach().cpu().numpy()
        if isinstance(data, list):
            return np.array(data)
        if isinstance(data, tuple):
            return [_to_numpy(item) for item in data if item is not None]
        return data

    converted_args = [_to_numpy(arg) for arg in args if arg is not None]
    converted_kwargs = {key: _to_numpy(val) for key, val in kwargs.items() if val is not None}

    if "dim" in converted_kwargs:
        converted_kwargs["axis"] = converted_kwargs.pop("dim")

    return return_fn(*converted_args, **converted_kwargs)

def torch_fn(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        is_torch_input = is_torch(*args, **kwargs)
        converted_args, converted_kwargs = to_torch(*args, return_fn=return_always, **kwargs)
        results = func(*converted_args, **converted_kwargs)
        return to_numpy(results, return_fn=return_if)[0] if not is_torch_input else results
    return wrapper

def numpy_fn(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        is_torch_input = is_torch(*args, **kwargs)
        device = "cuda" if is_cuda(*args, **kwargs) else "cpu"
        converted_args, converted_kwargs = to_numpy(*args, return_fn=return_always, **kwargs)
        results = func(*converted_args, **converted_kwargs)
        return results if not is_torch_input else to_torch(results, return_fn=return_if, device=device)[0][0]
    return wrapper

def batch_fn(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(x: Any, *args: Any, **kwargs: Any) -> Any:
        batch_size = int(kwargs.pop("batch_size", 4))
        if len(x) <= batch_size:
            return func(x, *args, **kwargs, batch_size=batch_size)
        n_batches = (len(x) + batch_size - 1) // batch_size
        results = []
        for i_batch in tqdm(range(n_batches)):
            start = i_batch * batch_size
            end = min((i_batch + 1) * batch_size, len(x))
            batch_result = func(x[start:end], *args, **kwargs, batch_size=batch_size)
            if isinstance(batch_result, torch.Tensor):
                batch_result = batch_result.cpu()
            elif isinstance(batch_result, tuple):
                batch_result = tuple(val.cpu() if isinstance(val, torch.Tensor) else val for val in batch_result)
            results.append(batch_result)
        if isinstance(results[0], tuple):
            n_vars = len(results[0])
            combined_results = [torch.vstack([res[i_var] for res in results]) for i_var in range(n_vars)]
            return tuple(combined_results)
        elif isinstance(results[0], torch.Tensor):
            return torch.vstack(results)
        else:
            return sum(results, [])
    return wrapper

def pandas_fn(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        is_torch_input = is_torch(*args, **kwargs)
        device = "cuda" if is_cuda(*args, **kwargs) else "cpu"
        def to_pandas(data: Any) -> pd.DataFrame:
            if isinstance(data, pd.DataFrame):
                return data
            elif isinstance(data, pd.Series):
                return pd.DataFrame(data)
            elif isinstance(data, (np.ndarray, list)):
                return pd.DataFrame(data)
            elif isinstance(data, torch.Tensor):
                return pd.DataFrame(data.detach().cpu().numpy())
            else:
                return pd.DataFrame([data])
        converted_args = [to_pandas(arg) for arg in args]
        converted_kwargs = {key: to_pandas(val) for key, val in kwargs.items()}
        results = func(*converted_args, **converted_kwargs)
        if is_torch_input:
            return to_torch(results, return_fn=return_if, device=device)[0]
        elif isinstance(results, (pd.DataFrame, pd.Series)):
            return results
        else:
            return to_numpy(results, return_fn=return_if)[0]
    return wrapper

if __name__ == "__main__":
    import scipy
    import torch.nn.functional as F

    @torch_fn
    def torch_softmax(*args: Any, **kwargs: Any) -> torch.Tensor:
        return F.softmax(*args, **kwargs)

    @numpy_fn
    def numpy_softmax(*args: Any, **kwargs: Any) -> np.ndarray:
        return scipy.special.softmax(*args, **kwargs)

    @pandas_fn
    def pandas_mean(*args: Any, **kwargs: Any) -> Union[pd.Series, float]:
        return pd.DataFrame(*args, **kwargs).mean()

    def custom_print(data: Any) -> None:
        print(type(data), data)

    test_data = [1, 2, 3]
    test_list = test_data
    test_tensor = torch.tensor(test_data).float()
    test_tensor_cuda = torch.tensor(test_data).float().cuda()
    test_array = np.array(test_data)
    test_df = pd.DataFrame({"col1": test_data})

    print("Testing torch_fn:")
    custom_print(torch_softmax(test_list, dim=-1))
    custom_print(torch_softmax(test_array, dim=-1))
    custom_print(torch_softmax(test_df, dim=-1))
    custom_print(torch_softmax(test_tensor, dim=-1))
    custom_print(torch_softmax(test_tensor_cuda, dim=-1))

    print("\nTesting numpy_fn:")
    custom_print(numpy_softmax(test_list, axis=-1))
    custom_print(numpy_softmax(test_array, axis=-1))
    custom_print(numpy_softmax(test_df, axis=-1))
    custom_print(numpy_softmax(test_tensor, axis=-1))
    custom_print(numpy_softmax(test_tensor_cuda, axis=-1))

    print("\nTesting pandas_fn:")
    custom_print(pandas_mean(test_list))
    custom_print(pandas_mean(test_array))
    custom_print(pandas_mean(test_df))
    custom_print(pandas_mean(test_tensor))
    custom_print(pandas_mean(test_tensor_cuda))



import functools
import warnings


def deprecated(reason=None):
    """
    A decorator to mark functions as deprecated. It will result in a warning being emitted
    when the function is used.

    Args:
        reason (str): A human-readable string explaining why this function was deprecated.
    """

    def decorator(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.warn(
                f"{func.__name__} is deprecated: {reason}",
                DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return new_func

    return decorator



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-12 19:54:42 (ywatanabe)"

from . import pac
from ._differential_bandpass_filters import (
    build_bandpass_filters,
    init_bandpass_filters,
)
from ._ensure_even_len import ensure_even_len
from ._zero_pad import zero_pad
from .filter import design_filter, plot_filter_responses



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-13 18:04:25 (ywatanabe)"
# /data/gpfs/projects/punim2354/ywatanabe/mngs_repo/src/mngs/general/_check_host.py

def check_host(keyword):
    import mngs
    return keyword in mngs.sh('echo $(hostname)', verbose=False)

is_host = check_host


def verify_host(keyword):
    import mngs
    import sys

    if is_host(keyword):
        print(f"Host verification successed for keyword: {keyword}")
        return
    else:
        print(f"Host verification failed for keyword: {keyword}")
        sys.exit(1)

if __name__ == '__main__':
    # check_host("ywata")
    verify_host("titan")
    verify_host("ywata")
    verify_host("crest")



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-30 12:11:01 (ywatanabe)"
# /mnt/ssd/ripple-wm-code/scripts/externals/mngs/src/mngs/dsp/_time.py

import mngs
import numpy as np


def time(start_sec, end_sec, fs):
    # return np.linspace(start_sec, end_sec, (end_sec - start_sec) * fs)
    return mngs.gen.float_linspace(
        start_sec, end_sec, (end_sec - start_sec) * fs
    )


def main():
    out = time(10, 15, 256)
    print(out)


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt
    import mngs

    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()
    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



import torch


def l1(model, lambda_l1=0.01):
    lambda_l1 = torch.tensor(lambda_l1)
    l1 = torch.tensor(0.0).cuda()
    for param in model.parameters(): # fixme; is this OK?
        l1 += torch.abs(param).sum()
    return l1


def l2(model, lambda_l2=0.01):
    lambda_l2 = torch.tensor(lambda_l2)
    l2 = torch.tensor(0.0).cuda()
    for param in model.parameters(): # fixme; is this OK?
        l2 += torch.norm(param).sum()
    return l2

def elastic(model, alpha=1.0, l1_ratio=0.5):
    assert 0 =< l1_ratio =< 1
    
    L1 = l1(model)
    L2 = l2(model)

    return alpha * (l1_ratio * L1 + (1 - l1_ratio) * L2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-24 13:13:33 (ywatanabe)"
# /mnt/ssd/mngs_repo/src/mngs/db/_inspect.py

import sqlite3
from typing import List, Tuple, Any, Optional, Dict
import os
import mngs


class Inspector:
    def __init__(self, db_path: str):
        if not os.path.exists(db_path):
            raise FileNotFoundError(f"Database file not found: {db_path}")
        self.db_path = db_path

    def get_table_names(self) -> List[str]:
        """Retrieves all table names from the database.

        Returns:
            List[str]: List of table names
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                "SELECT name FROM sqlite_master WHERE type='table';"
            )
            return [table[0] for table in cursor.fetchall()]

    def get_table_info(
        self, table_name: str
    ) -> List[Tuple[int, str, str, int, Any, int, str]]:
        """Retrieves table structure information.

        Args:
            table_name (str): Name of the table

        Returns:
            List[Tuple[int, str, str, int, Any, int, str]]: List of column information tuples
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(f"PRAGMA table_info({table_name})")
            columns = cursor.fetchall()

            cursor.execute(f"PRAGMA index_list({table_name})")
            indexes = cursor.fetchall()
            pk_columns = []
            for idx in indexes:
                if idx[2] == 1:  # Is primary key
                    cursor.execute(f"PRAGMA index_info({idx[1]})")
                    pk_columns.extend([info[2] for info in cursor.fetchall()])

            enhanced_columns = []
            for col in columns:
                constraints = []
                if col[1] in pk_columns:
                    constraints.append("PRIMARY KEY")
                if col[3] == 1:
                    constraints.append("NOT NULL")
                enhanced_columns.append(col + (" ".join(constraints),))

            return enhanced_columns

    def get_sample_data(
        self, table_name: str, limit: int = 5
    ) -> Tuple[List[str], List[Tuple], int]:
        """Retrieves sample data from the specified table.

        Args:
            table_name (str): Name of the table
            limit (int, optional): Number of rows to retrieve. Defaults to 5.

        Returns:
            Tuple[List[str], List[Tuple], int]: Column names, sample data rows, and total row count
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(f"SELECT * FROM {table_name} LIMIT {limit}")
            columns = [description[0] for description in cursor.description]
            sample_data = cursor.fetchall()

            cursor.execute(f"SELECT COUNT(*) FROM {table_name}")
            total_rows = cursor.fetchone()[0]

            return columns, sample_data, total_rows

    def inspect(
            self, table_names: Optional[List[str]] = None, verbose=True,
    ) -> List[Dict[str, Any]]:
        import pandas as pd

        if table_names is None:
            table_names = self.get_table_names()

        data_tables = []
        for table_name in table_names:
            columns = self.get_table_info(table_name)
            column_names, rows, total_rows = self.get_sample_data(table_name)

            meta = {}
            meta["table_name"] = table_name
            meta["n_total_rows"] = total_rows


            sample_data = pd.DataFrame(
                [
                    {
                        col: (
                            str(value)
                            if not isinstance(value, bytes)
                            else "<BLOB>"
                        )
                        for col, value in zip(column_names, row)
                    }
                    for row in rows
                ]
            )

            for k, v in meta.items():
                sample_data[k] = v

            sample_data = sample_data.set_index(["table_name", "n_total_rows"])

            data_tables.append(sample_data)

        # if len(data_tables) == 1:
        #     return data_tables[0]
        # else:
        #     return tuple(data_tables)
        return data_tables


def inspect(lpath_db: str, table_names: Optional[List[str]] = None, verbose: bool = True) -> None:
    """
    Inspects the specified SQLite database.

    Example:
    >>> inspect('path/to/database.db')
    >>> inspect('path/to/database.db', ['table1', 'table2'])

    Args:
        lpath_db (str): Path to the SQLite database file
        table_names (Optional[List[str]], optional): List of table names to inspect.
            If None, inspects all tables. Defaults to None.
    """
    inspector = Inspector(lpath_db)
    overviews_tables = inspector.inspect(table_names, verbose=verbose)
    if verbose:
        for dd in overviews_tables:
            print(f"\n{dd}\n")
    return overviews_tables


# python -c "import mngs; mngs.db.inspect(\"./data/db_all/Patient_23_005.db\")"
# python -c "import mngs; mngs.db.inspect(\"./data/db_all/Patient_23_005.db\", table_names=[\"eeg_data_reindexed\"])"
# python -c "import mngs; mngs.db.inspect(\"./data/db_all/Patient_23_005.db\", table_names=[\"eeg_data\"])"
# python -c "import mngs; mngs.db.inspect(\"./data/db_all/Patient_23_005.db\", table_names=[\"sqlite_sequence\"])"



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-24 15:05:34"
# Author: Yusuke Watanabe (ywata1989@gmail.com)


"""
This script does XYZ.
"""


"""
Imports
"""
import sys

import matplotlib.pyplot as plt
import mngs

"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def title_case(text):
    """
    Converts a string to title case while keeping certain prepositions, conjunctions, and articles in lowercase,
    and ensuring words detected as potential acronyms (all uppercase) are fully capitalized.

    Parameters:
    - text (str): The text to convert to title case.

    Returns:
    - str: The converted text in title case with certain words in lowercase and potential acronyms fully capitalized.

    Examples:
    --------
        print(title_case("welcome to the world of ai and using CPUs for gaming"))  # Welcome to the World of AI and Using CPUs for Gaming
    """
    # List of words to keep in lowercase
    lowercase_words = [
        "a",
        "an",
        "the",
        "and",
        "but",
        "or",
        "nor",
        "at",
        "by",
        "to",
        "in",
        "with",
        "of",
        "on",
    ]

    words = text.split()
    final_words = []
    for word in words:
        # Check if the word is fully in uppercase and more than one character, suggesting an acronym
        if word.isupper() and len(word) > 1:
            final_words.append(word)
        elif word.lower() in lowercase_words:
            final_words.append(word.lower())
        else:
            final_words.append(word.capitalize())
    return " ".join(final_words)


def main():
    # Example usage:
    text = "welcome to the world of ai and using CPUs for gaming"
    print(title_case(text))


if __name__ == "__main__":
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-06 09:00:45 (ywatanabe)"

# import os
# import shutil

# def mv(src, tgt):
#     successful = True
#     os.makedirs(tgt, exist_ok=True)

#     if os.path.isdir(src):
#         # Iterate over the items in the directory
#         for item in os.listdir(src):
#             item_path = os.path.join(src, item)
#             # Check if the item is a file
#             if os.path.isfile(item_path):
#                 try:
#                     shutil.move(item_path, tgt)
#                     print(f"\nMoved file from {item_path} to {tgt}")
#                 except OSError as e:
#                     print(f"\nError: {e}")
#                     successful = False
#             else:
#                 print(f"\nSkipped directory {item_path}")
#     else:
#         # If src is a file, just move it
#         try:
#             shutil.move(src, tgt)
#             print(f"\nMoved from {src} to {tgt}")
#         except OSError as e:
#             print(f"\nError: {e}")
#             successful = False

#     return successful


def mv(src, tgt):
    import os
    import shutil

    successful = True
    os.makedirs(tgt, exist_ok=True)

    try:
        shutil.move(src, tgt)
        print(f"\nMoved from {src} to {tgt}")
    except OSError as e:
        print(f"\nError: {e}")
        successful = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-13 16:39:20 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/general/utils/_weather.py

"""This script does XYZ."""

"""Imports"""
import importlib
import logging
import os
import re
import sys
import warnings
from glob import glob
from pprint import pprint

import matplotlib
import matplotlib.pyplot as plt
import mngs
import numpy as np
import pandas as pd
import seaborn as sns
import torch
import torch.nn as nn
import torch.nn.functional as F
import xarray as xr
from icecream import ic
from natsort import natsorted
from tqdm import tqdm

try:
    from scripts import utils
except:
    pass

"""Config"""
# CONFIG = mngs.gen.load_configs()

"""Functions & Classes"""
import time

import requests
from plyer import notification

API_KEY = "your_api_key"
CITY = "your_city"
API_URL = (
    f"http://api.openweathermap.org/data/2.5/weather?q={CITY}&appid={API_KEY}"
)


def check_rain():
    response = requests.get(API_URL)
    data = response.json()
    if "rain" in data:
        notify_rain()


def notify_rain():
    notification.notify(
        title="Rain Alert",
        message="It's starting to rain in your area!",
        timeout=10,
    )


while True:
    check_rain()
    time.sleep(300)  # Check every 5 minutes

if __name__ == "__main__":
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False, agg=True
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-08-25 10:19:47 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/general/_dict2str.py


def dict2str(dictionary, delimiter="_"):
    """
    Convert a dictionary to a string representation.

    Example
    -------
    input_dict = {'a': 1, 'b': 2, 'c': 3}
    result = dict2str(input_dict)
    print(result)  # Output: a-1_b-2_c-3

    Parameters
    ----------
    dictionary : dict
        The input dictionary to be converted.
    delimiter : str, optional
        The separator between key-value pairs (default is "_").

    Returns
    -------
    str
        A string representation of the input dictionary.
    """
    return delimiter.join(
        f"{key}-{value}" for key, value in dictionary.items()
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-20 02:17:10 (ywatanabe)"
# /data/gpfs/projects/punim2354/ywatanabe/mngs_repo/src/mngs/db/_delete_duplicates_clean.py

"""
Functionality:
    - Deletes duplicate entries from an SQLite database table
Input:
    - SQLite database file path, table name, columns to consider for duplicates
Output:
    - Updated SQLite database with duplicates removed
Prerequisites:
    - sqlite3, pandas, tqdm, mngs
"""

import sqlite3
import pandas as pd
from typing import Union, List, Optional, Tuple
from tqdm import tqdm
import mngs

def _sort_db(
    cursor: sqlite3.Cursor,
    table_name: str,
    columns: List[str]
) -> None:
    """
    Sorts the database table based on the specified columns.

    Parameters
    ----------
    cursor : sqlite3.Cursor
        The cursor object for executing SQL commands.
    table_name : str
        The name of the table to be sorted.
    columns : List[str]
        The list of column names to sort by, in order of priority.

    Example
    -------
    >>> conn = sqlite3.connect('example.db')
    >>> cursor = conn.cursor()
    >>> _sort_db(cursor, 'my_table', ['column1', 'column2'])
    >>> conn.commit()
    >>> conn.close()
    """
    columns_str = ', '.join(columns)
    temp_table = f"{table_name}_temp"

    cursor.execute(f"CREATE TABLE {temp_table} AS SELECT * FROM {table_name} ORDER BY {columns_str}")
    cursor.execute(f"DROP TABLE {table_name}")
    cursor.execute(f"ALTER TABLE {temp_table} RENAME TO {table_name}")


def _determine_columns(
    cursor: sqlite3.Cursor,
    table_name: str,
    columns: Union[str, List[str]],
    include_blob: bool,
) -> List[str]:
    cursor.execute(f"PRAGMA table_info({table_name})")
    table_info = cursor.fetchall()
    all_columns = [col[1] for col in table_info]
    column_types = {col[1]: col[2] for col in table_info}

    if columns == "all":
        columns = (
            all_columns
            if include_blob
            else [
                col
                for col in all_columns
                if column_types[col].lower() != "blob"
            ]
        )
    elif isinstance(columns, str):
        columns = [columns]

    columns_str = ", ".join(columns)
    print(f"Columns considered for duplicates: {columns_str}")

    return columns

def _fetch_as_df(
    cursor: sqlite3.Cursor, columns: List[str], table_name: str
) -> pd.DataFrame:
    print("\nFetching all database entries...")
    columns_str = ", ".join(columns)
    query = f"SELECT {columns_str} FROM {table_name}"
    cursor.execute(query)
    df_entries = cursor.fetchall()
    return pd.DataFrame(df_entries, columns=columns)

def _find_duplicated(df: pd.DataFrame) -> pd.DataFrame:
    df_duplicated = df[df.duplicated(keep="first")].copy()
    duplication_rate = len(df_duplicated) / (len(df) - len(df_duplicated))
    print(
        f"\n{100*duplication_rate:.2f}% of data was duplicated. Cleaning up..."
    )
    print(f"\nOriginal entries:\n{df.head()}")
    print(f"\nDuplicated entries:\n{df_duplicated.head()}")
    return df_duplicated

def verify_duplicated_index(
    cursor: sqlite3.Cursor, duplicated_row: pd.Series, table_name: str, dry_run: bool
) -> Tuple[str, bool]:
    """Check if entry to delete is the one intended"""
    columns = list(duplicated_row.index)
    columns_str = ", ".join(columns)

    where_conditions = " AND ".join([f"{col} = ?" for col in columns])
    select_query = f"""
        SELECT {columns_str}
        FROM {table_name}
        WHERE {where_conditions}
    """
    cursor.execute(select_query, tuple(duplicated_row))
    entries = cursor.fetchall()

    is_verified = len(entries) >= 1

    if dry_run:
        print(f"Expected duplicate entry: {tuple(duplicated_row)}")
        print(f"Found entries: {entries}")
        print(f"Verification {'succeeded' if is_verified else 'failed'}")

    return select_query, is_verified

def _delete_entry(
    cursor: sqlite3.Cursor,
    duplicated_row: pd.Series,
    table_name: str,
    dry_run: bool = True,
) -> None:
    select_query, is_verified = verify_duplicated_index(cursor, duplicated_row, table_name, dry_run)
    if is_verified:
        delete_query = select_query.replace("SELECT", "DELETE")
        if dry_run:
            print(f"[DRY RUN] Would delete entry:\n{duplicated_row}")
        else:
            cursor.execute(delete_query, tuple(duplicated_row))
            print(f"Deleted entry:\n{duplicated_row}")
    else:
        print(f"Skipping entry (not found or already deleted):\n{duplicated_row}")


# def delete_duplicates(
#     lpath_db: str,
#     table_name: str,
#     columns: Union[str, List[str]] = "all",
#     include_blob: bool = False,
#     batch_size: int = 1000,
#     reindex: bool = False,
#     sort: bool = False,
#     dry_run: bool = True,
# ) -> Tuple[Optional[pd.DataFrame], Optional[pd.DataFrame]]:
#     """
#     Delete duplicate entries from an SQLite database table.

#     Parameters
#     ----------
#     lpath_db : str
#         Path to the SQLite database file.
#     table_name : str
#         Name of the table to remove duplicates from.
#     columns : Union[str, List[str]], optional
#         Columns to consider when identifying duplicates. Default is "all".
#     include_blob : bool, optional
#         Whether to include BLOB columns when considering duplicates. Default is False.
#     batch_size : int, optional
#         Number of rows to process in each batch. Default is 1000.
#     reindex : bool, optional
#         Whether to reindex the table after deletion. Default is False.
#     dry_run : bool, optional
#         If True, simulates the deletion without actually modifying the database. Default is True.

#     Returns
#     -------
#     Tuple[Optional[pd.DataFrame], Optional[pd.DataFrame]]
#         A tuple containing:
#         - DataFrame of all entries after deletion process
#         - DataFrame of remaining duplicates if any, None otherwise
#     """
#     try:
#         conn = sqlite3.connect(lpath_db)
#         cursor = conn.cursor()

#         columns = _determine_columns(cursor, table_name, columns, include_blob)

#         if sort:
#             _sort_db(cursor, table_name, columns)

#         df_orig = _fetch_as_df(cursor, columns, table_name)
#         duplicates = _find_duplicated(df_orig)

#         if duplicates.empty:
#             print("Congratulations. Database is clean.")
#             return df_orig, None

#         columns_str = ", ".join(columns)
#         where_conditions = " AND ".join([f"{col} = ?" for col in columns])
#         delete_query = f"""
#             DELETE FROM {table_name}
#             WHERE {where_conditions}
#         """

#         for start in tqdm(range(0, len(duplicates), batch_size)):
#             batch = duplicates.iloc[start:start+batch_size]
#             batch_values = batch.values.tolist()

#             if dry_run:
#                 print(f"[DRY RUN] Would delete {len(batch)} entries")
#             else:
#                 cursor.executemany(delete_query, batch_values)
#                 conn.commit()

#         if not dry_run:
#             conn.commit()

#             if reindex:
#                 print("Reindexing the table...")
#                 cursor.execute(f"REINDEX {table_name}")
#                 conn.commit()

#         df_after = _fetch_as_df(cursor, columns, table_name)
#         remaining_duplicates = _find_duplicated(df_after)

#         if remaining_duplicates.empty:
#             print("All duplicates successfully removed.")
#             return df_after, None
#         else:
#             print(f"Warning: {len(remaining_duplicates)} duplicates still remain.\n{remaining_duplicates}")
#             return df_after, remaining_duplicates

#     except Exception as error:
#         print(f"An error occurred: {error}")
#         return None, None

#     finally:
#         conn.close()

# def delete_duplicates(
#     lpath_db: str,
#     table_name: str,
#     columns: Union[str, List[str]] = "all",
#     include_blob: bool = False,
#     batch_size: int = 1000,
#     chunk_size: int = 100_000,
#     reindex: bool = False,
#     sort: bool = False,
#     dry_run: bool = True,
# ) -> Tuple[Optional[int], Optional[int]]:
#     try:
#         conn = sqlite3.connect(lpath_db)
#         cursor = conn.cursor()

#         columns = _determine_columns(cursor, table_name, columns, include_blob)

#         if sort:
#             _sort_db(cursor, table_name, columns)

#         columns_str = ", ".join(columns)
#         where_conditions = " AND ".join([f"{col} = ?" for col in columns])
#         delete_query = f"""
#             DELETE FROM {table_name}
#             WHERE {where_conditions}
#         """

#         total_rows = cursor.execute(f"SELECT COUNT(*) FROM {table_name}").fetchone()[0]
#         total_deleted = 0
#         total_duplicates = 0

#         for offset in tqdm(range(0, total_rows, chunk_size)):
#             chunk_query = f"""
#                 SELECT {columns_str}
#                 FROM {table_name}
#                 LIMIT {chunk_size} OFFSET {offset}
#             """
#             df_chunk = pd.read_sql_query(chunk_query, conn)
#             duplicates = _find_duplicated(df_chunk)
#             total_duplicates += len(duplicates)

#             if duplicates.empty:
#                 continue

#             for start in range(0, len(duplicates), batch_size):
#                 batch = duplicates.iloc[start:start+batch_size]
#                 batch_values = batch.values.tolist()

#                 if dry_run:
#                     print(f"[DRY RUN] Would delete {len(batch)} entries")
#                 else:
#                     cursor.executemany(delete_query, batch_values)
#                     conn.commit()
#                     total_deleted += len(batch)

#         if not dry_run:
#             if reindex:
#                 print("Reindexing the table...")
#                 cursor.execute(f"REINDEX {table_name}")
#                 conn.commit()

#         print(f"Total duplicates found: {total_duplicates}")
#         print(f"Total entries deleted: {total_deleted}")

#         return total_duplicates, total_deleted

#     except Exception as error:
#         print(f"An error occurred: {error}")
#         return None, None

#     finally:
#         conn.close()


def delete_duplicates(
    lpath_db: str,
    table_name: str,
    columns: Union[str, List[str]] = "all",
    include_blob: bool = False,
    chunk_size: int = 10_000,
    dry_run: bool = True,
) -> Tuple[Optional[int], Optional[int]]:
    try:
        conn = sqlite3.connect(lpath_db)
        cursor = conn.cursor()

        # Vacuum the database to free up space
        if not dry_run:
            cursor.execute("VACUUM")
            conn.commit()

        columns = _determine_columns(cursor, table_name, columns, include_blob)
        columns_str = ", ".join(columns)

        # Create a temporary table to store unique rows
        temp_table = f"{table_name}_temp"
        cursor.execute(f"CREATE TABLE {temp_table} AS SELECT DISTINCT {columns_str} FROM {table_name} LIMIT 0")

        # Process in small chunks
        offset = 0
        total_processed = 0
        total_unique = 0

        while True:
            chunk_query = f"""
                INSERT OR IGNORE INTO {temp_table}
                SELECT DISTINCT {columns_str}
                FROM {table_name}
                LIMIT {chunk_size} OFFSET {offset}
            """

            if dry_run:
                print(f"[DRY RUN] Would execute: {chunk_query}")
            else:
                cursor.execute(chunk_query)
                conn.commit()

            rows_affected = cursor.rowcount
            if rows_affected == 0:
                break

            total_processed += chunk_size
            total_unique += rows_affected
            offset += chunk_size

            print(f"Processed {total_processed} rows, {total_unique} unique")

        total_duplicates = total_processed - total_unique

        if not dry_run:
            # Replace original table with the deduplicated one
            cursor.execute(f"DROP TABLE {table_name}")
            cursor.execute(f"ALTER TABLE {temp_table} RENAME TO {table_name}")
            conn.commit()

        print(f"Total rows processed: {total_processed}")
        print(f"Total unique rows: {total_unique}")
        print(f"Total duplicates removed: {total_duplicates}")

        return total_processed, total_duplicates

    except Exception as error:
        print(f"An error occurred: {error}")
        return None, None

    finally:
        conn.close()



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-13 20:10:40 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/_gen_AI/_ChatGPT.py


"""Imports"""
import os
import sys

import anthropic
import matplotlib.pyplot as plt
import mngs

from ._BaseGenAI import BaseGenAI

"""Functions & Classes"""


class Anthropic(BaseGenAI):
    def __init__(
        self,
        system_setting="",
        api_key=os.getenv("ANTHROPIC_API_KEY"),
        model="claude-3-opus-20240229",
        stream=False,
        seed=None,
        n_keep=1,
        temperature=1.0,
        chat_history=None,
        max_tokens=4096,
    ):
        super().__init__(
            system_setting=system_setting,
            model=model,
            api_key=api_key,
            stream=stream,
            n_keep=n_keep,
            temperature=temperature,
            provider="Anthropic",
            chat_history=chat_history,
            max_tokens=max_tokens,
        )

    def _init_client(
        self,
    ):
        client = anthropic.Anthropic(
            api_key=self.api_key,
        )
        return client

    def _api_call_static(
        self,
    ):
        output = self.client.messages.create(
            model=self.model,
            max_tokens=self.max_tokens,
            messages=self.history,
            temperature=self.temperature,
        )
        out_text = output.content[0].text

        self.input_tokens += output.usage.input_tokens
        self.output_tokens += output.usage.output_tokens

        return out_text

    def _api_call_stream(self):

        with self.client.messages.stream(
            model=self.model,
            max_tokens=self.max_tokens,
            messages=self.history,
            temperature=self.temperature,
        ) as stream:
            for chunk in stream:

                try:
                    self.input_tokens += chunk.message.usage.input_tokens
                    self.output_tokens += chunk.message.usage.output_tokens
                except:
                    pass
                if chunk.type == "content_block_delta":
                    yield chunk.delta.text

    # def _get_available_models(self):
    #     return [
    #         "claude-3-5-sonnet-20240620",
    #         "claude-3-opus-20240229",
    #         "claude-3-sonnet-20240229",
    #         "claude-3-haiku-20240307",
    #     ]


def main():
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-23 19:11:33"
# Author: Yusuke Watanabe (ywata1989@gmail.com)


"""
This script does XYZ.
"""


"""
Imports
"""
import os
import sys

import matplotlib.pyplot as plt
import mngs
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F

"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""
import time
from multiprocessing import Process, Queue


def timeout(seconds=10, error_message="Timeout"):
    def decorator(func):
        def wrapper(*args, **kwargs):
            def queue_wrapper(queue, args, kwargs):
                result = func(*args, **kwargs)
                queue.put(result)

            queue = Queue()
            args_for_process = (queue, args, kwargs)
            process = Process(target=queue_wrapper, args=args_for_process)
            process.start()
            process.join(timeout=seconds)

            if process.is_alive():
                process.terminate()
                raise TimeoutError(error_message)
            else:
                return queue.get()

        return wrapper

    return decorator


def main():
    # Example usage
    @timeout(seconds=3, error_message="Function call timed out")
    def long_running_function(x):
        time.sleep(4)  # Simulate a long-running operation
        return x

    try:
        result = long_running_function(10)
        print(f"Result: {result}")
    except TimeoutError as e:
        print(e)


if __name__ == "__main__":
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-19 11:30:45 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/general/system_ops/_notify.py

"""This script does XYZ."""

import inspect
import os
import pwd
import socket
import subprocess
import sys

import mngs

from ._email import send_gmail
import warnings

def get_username():
    try:
        return pwd.getpwuid(os.getuid()).pw_name
    except:
        return os.getenv("USER") or os.getenv("LOGNAME") or "unknown"


def get_hostname():
    return socket.gethostname()


def get_git_branch():
    try:
        branch = (
            subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                cwd=mngs.__path__[0],
                stderr=subprocess.DEVNULL,
            )
            .decode()
            .strip()
        )
        return branch
    except Exception as e:
        return "main"


def gen_footer(sender, script_name, mngs, branch):
    return f"""

{'-'*30}
Sent via
- Host: {sender}
- Script: {script_name}
- Source: mngs v{mngs.__version__} (github.com/ywatanabe1989/mngs/blob/{branch}/src/mngs/general/system_ops/_notify.py)
{'-'*30}"""


# This is an automated system notification. If received outside working hours, please disregard.


def notify(
    subject="",
    message=":)",
    ID="auto",
    sender_name=None,
    recipient_email=None,
    cc=None,
    attachment_paths=None,
    verbose=False,
):
    try:
        message = str(message)
    except Exception as e:
        warnings.warn(str(e))

    FAKE_PYTHON_SCRIPT_NAME = "$ python -c ..."
    sender_gmail = os.getenv("MNGS_SENDER_GMAIL")
    sender_password = os.getenv("MNGS_SENDER_GMAIL_PASSWORD")
    recipient_email = recipient_email or os.getenv("MNGS_RECIPIENT_GMAIL")

    if sys.argv[0]:
        script_name = os.path.basename(sys.argv[0])
    else:
        frames = inspect.stack()
        script_name = (
            os.path.basename(frames[-1].filename) if frames else "(Not found)"
        )
    if (script_name == "-c") or (not script_name.endswith(".py")):
        script_name = FAKE_PYTHON_SCRIPT_NAME

    sender = f"{get_username()}@{get_hostname()}"
    branch = get_git_branch()
    footer = gen_footer(sender, script_name, mngs, branch)

    full_message = script_name + "\n\n" + message + "\n\n" + footer
    full_subject = (
        f"{script_name}—{subject}"
        if subject and (script_name != FAKE_PYTHON_SCRIPT_NAME)
        else f"{subject}"
    )

    if sender_gmail is None or sender_password is None:
        print(
            f"""
        Please set environmental variables to use this function (f"{inspect.stack()[0][3]}"):\n\n
        $ export MNGS_SENDER_GMAIL="mngs.notification@gmail.com"
        $ export MNGS_SENDER_GMAIL_PASSWORD="YOUR_APP_PASSWORD"
        $ export MNGS_RECIPIENT_GMAIL="YOUR_GMAIL_ADDRESS"
        """
        )

    send_gmail(
        sender_gmail,
        sender_password,
        recipient_email,
        full_subject,
        full_message,
        sender_name=sender_name,
        cc=cc,
        ID=ID,
        attachment_paths=attachment_paths,
        verbose=verbose,
    )


if __name__ == "__main__":
    notify(verbose=True)

    # python -c "import mngs; mngs.gen.notify()"


# # Example in shell
# #!/bin/bash
# # /home/ywatanabe/.dotfiles/.bin/notify
# # Author: ywatanabe (ywatanabe@alumni.u-tokyo.ac.jp)
# # Date: $(date +"%Y-%m-%d-%H-%M")

# # LOG_FILE="${0%.sh}.log"

# usage() {
#     echo "Usage: $0 [-s|--subject <subject>] [-m|--message <message>] [-r|--recipient-name <name>] [-t|--to <email>] [-c|--cc <email>] [-h|--help]"
#     echo "Options:"
#     echo "  -s, --subject   Subject of the notification (default: 'Subject')"
#     echo "  -m, --message   Message body of the notification (default: 'Message')"
#     echo "  -t, --to        The email address of the recipient"
#     echo "  -c, --cc        CC email address(es) (can be used multiple times)"
#     echo "  -h, --help      Display this help message"
#     echo
#     echo "Example:"
#     echo "  $0 -s \"About the Project A\" -m \"Hi, ...\" -r \"John\" -t \"john@example.com\" -c \"cc1@example.com\" -c \"cc2@example.com\""
#     echo "  $0 -s \"Notification\" -m \"This is a notification from ...\" -r \"Team\" -t \"team@example.com\""
#     exit 1
# }

# main() {
#     subject="Subject"
#     message="Message"
#     recipient_email=""
#     cc_addresses=()

#     while [[ $# -gt 0 ]]; do
#         case $1 in
#             -s|--subject)
#                 shift
#                 subject=""
#                 while [[ $# -gt 0 && ! $1 =~ ^- ]]; do
#                     subject+="$1 "
#                     shift
#                 done
#                 subject=${subject% }
#                 ;;
#             -m|--message)
#                 shift
#                 message=""
#                 while [[ $# -gt 0 && ! $1 =~ ^- ]]; do
#                     message+="$1 "
#                     shift
#                 done
#                 message=${message% }
#                 ;;
#             -t|--to)
#                 shift
#                 recipient_email="$1"
#                 shift
#                 ;;
#             -c|--cc)
#                 shift
#                 cc_addresses+=("$1")
#                 shift
#                 ;;
#             -h|--help)
#                 usage
#                 ;;
#             *)
#                 echo "Unknown option: $1"
#                 usage
#                 ;;
#         esac
#     done

#     subject=$(echo "$subject" | sed "s/'/'\\\\''/g")
#     message=$(echo "$message" | sed "s/'/'\\\\''/g")
#     recipient_email=$(echo "$recipient_email" | sed "s/'/'\\\\''/g")
#     cc_string=$(IFS=,; echo "${cc_addresses[*]}" | sed "s/'/'\\\\''/g")

#     python -c "
# import mngs

# cc_list = [$(printf "'%s', " "${cc_addresses[@]}")]
# cc_list = [addr.strip() for addr in cc_list if addr.strip()]

# mngs.gen.notify(
#     subject='$subject',
#     message='$message',
#     ID=None,
#     recipient_email='$recipient_email',
#     cc=cc_list
# )
# "
# }

# main "$@"
# # { main "$@"; } 2>&1 | tee "$LOG_FILE"

# # EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-02 09:21:12 (ywatanabe)"

import torch
import torch.nn as nn
import torch.nn.functional as F
from mngs.general import numpy_fn, torch_fn


class Spectrogram(nn.Module):
    def __init__(
        self,
        sampling_rate,
        n_fft=256,
        hop_length=None,
        win_length=None,
        window="hann",
    ):
        super().__init__()
        self.sampling_rate = sampling_rate
        self.n_fft = n_fft
        self.hop_length = hop_length if hop_length is not None else n_fft // 4
        self.win_length = win_length if win_length is not None else n_fft
        if window == "hann":
            self.window = torch.hann_window(window_length=self.win_length)
        else:
            raise ValueError(
                "Unsupported window type. Extend this to support more window types."
            )

    def forward(self, x):
        """
        Computes the spectrogram for each channel in the input signal.

        Parameters:
        - signal (torch.Tensor): Input signal of shape (batch_size, n_chs, seq_len).

        Returns:
        - spectrograms (torch.Tensor): The computed spectrograms for each channel.
        """

        x = mngs.dsp.ensure_3d(x)

        batch_size, n_chs, seq_len = x.shape
        spectrograms = []

        for ch in range(n_chs):
            x_ch = x[:, ch, :].unsqueeze(
                1
            )  # Maintain expected input shape for stft
            spec = torch.stft(
                x_ch.squeeze(1),
                n_fft=self.n_fft,
                hop_length=self.hop_length,
                win_length=self.win_length,
                window=self.window.to(x.device),
                center=True,
                pad_mode="reflect",
                normalized=False,
                return_complex=True,
            )
            magnitude = torch.abs(spec).unsqueeze(1)  # Keep channel dimension
            spectrograms.append(magnitude)

        # Concatenate spectrograms along channel dimension
        spectrograms = torch.cat(spectrograms, dim=1)

        # Calculate frequencies (y-axis)
        freqs = torch.linspace(
            0, self.sampling_rate / 2, steps=self.n_fft // 2 + 1
        )

        # Calculate times (x-axis)
        # The number of frames can be computed from the size of the last dimension of the spectrogram
        n_frames = spectrograms.shape[-1]
        # Time of each frame in seconds, considering the hop length and sampling rate
        times_sec = torch.arange(0, n_frames) * (
            self.hop_length / self.sampling_rate
        )

        return spectrograms, freqs, times_sec


@torch_fn
def spectrograms(x, fs, cuda=False):
    return Spectrogram(fs)(x)


@torch_fn
def my_softmax(x, dim=-1):
    return F.softmax(x, dim=dim)


@torch_fn
def unbias(x, func="min", dim=-1, cuda=False):
    if func == "min":
        return x - x.min(dim=dim, keepdims=True)[0]
    if func == "mean":
        return x - x.mean(dim=dim, keepdims=True)[0]


@torch_fn
def normalize(x, axis=-1, amp=1.0, cuda=False):
    high = torch.abs(x.max(axis=axis, keepdims=True)[0])
    low = torch.abs(x.min(axis=axis, keepdims=True)[0])
    return amp * x / torch.maximum(high, low)


@torch_fn
def spectrograms(x, fs, dj=0.125, cuda=False):
    from wavelets_pytorch.transform import (
        WaveletTransformTorch,
    )  # PyTorch version

    dt = 1 / fs
    # dj = 0.125
    batch_size, n_chs, seq_len = x.shape

    x = x.cpu().numpy()

    # # Batch of signals to process
    # batch = np.array([batch_size * seq_len])

    # Initialize wavelet filter banks (scipy and torch implementation)
    # wa_scipy = WaveletTransform(dt, dj)
    wa_torch = WaveletTransformTorch(dt, dj, cuda=True)

    # Performing wavelet transform (and compute scalogram)
    # cwt_scipy = wa_scipy.cwt(batch)
    x = x[:, 0][:, np.newaxis]
    cwt_torch = wa_torch.cwt(x)

    return cwt_torch


if __name__ == "__main__":
    import mngs
    import seaborn as sns
    import torchaudio

    fs = 1024  # 128
    t_sec = 10
    x = mngs.dsp.np.demo_sig(t_sec=t_sec, fs=fs, type="ripple")

    normalize(unbias(x, cuda=True), cuda=True)

    # My implementtion
    ss = spectrograms(x, fs, cuda=True)
    fig, axes = plt.subplots(nrows=2)
    axes[0].plot(np.arange(x[0, 0]) / fs, x[0, 0])
    sns.heatmap(ss[0], ax=axes[1])
    plt.show()

    ss, ff, tt = spectrograms(x, fs, cuda=True)
    fig, axes = plt.subplots(nrows=2)
    axes[0].plot(np.arange(x[0, 0]) / fs, x[0, 0])
    sns.heatmap(ss[0], ax=axes[1])
    plt.show()

    # Torch Audio
    transform = torchaudio.transforms.Spectrogram(
        n_fft=16, normalized=True
    ).cuda()
    xx = torch.tensor(x).float().cuda()[0, 0]
    ss = transform(xx)
    sns.heatmap(ss.detach().cpu().numpy())

    plt.show()



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-29 22:30:22 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/web/_crawl.py


"""
This script does XYZ.
"""


"""
Imports
"""
import os
import re
import sys

import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import importlib

import mngs

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from icecream import ic
from natsort import natsorted
from glob import glob
from pprint import pprint
import warnings
import logging
from tqdm import tqdm
import xarray as xr

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def main():
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-15 18:10:23 (ywatanabe)"

import matplotlib.pyplot as plt
import mngs
import numpy as np


def configure_mpl(
    plt,
    # Fig Size
    fig_size_mm=(160, 100),
    fig_scale=1.0,
    # DPI
    dpi_display=100,
    dpi_save=300,
    # # Font Size
    # font_size_base=10,
    # font_size_title=10,
    # font_size_axis_label=10,
    # font_size_tick_label=8,
    # font_size_legend=8,
    # Hide spines
    hide_top_right_spines=True,
    # line
    line_width=0.5,
    # Color transparency
    alpha=0.9,
    # Whether to print configurations or not
    verbose=False,
    **kwargs,
):
    """
    Configures Matplotlib and Seaborn settings for publication-quality plots.
    For axis control, refer to the mngs.plt.ax module.

    Example:
        plt, CC = configure_mpl(plt)

        fig, ax = plt.subplots()
        x = np.linspace(0, 10, 100)
        for i_cc, cc_str in enumerate(cc):
            phase_shift = i_cc * np.pi / len(CC)
            y = np.sin(x + phase_shift)
            ax.plot(x, y, color=CC[cc_str], label=f"{cc_str}")
        ax.legend()
        plt.show()

    Parameters:
        plt (matplotlib.pyplot):
            Matplotlib pyplot module.

        fig_size_mm (tuple of int, optional):
            Figure width and height in millimeters. Defaults to (160, 100).

        fig_scale (float, optional):
            Scaling factor for the figure size. Defaults to 1.0.

        dpi_display (int, optional):
            Display resolution in dots per inch. Defaults to 100.

        dpi_save (int, optional):
            Resolution for saved figures in dots per inch. Defaults to 300.

        # font_size_title (int, optional):
        #     Font size for titles. Defaults to 8.

        # font_size_axis_label (int, optional):
        #     Font size for axis labels. Defaults to 8.

        # font_size_tick_label (int, optional):
        #     Font size for tick labels. Defaults to 7.

        # font_size_legend (int, optional):
        #     Font size for legend text. Defaults to 6.

        hide_top_right_spines (bool, optional):
            If True, hides the top and right spines of the plot. Defaults to True.

        alpha (float, optional):
            Color transparency. Defaults to 0.75.

        verbose (bool, optional):
            If True, prints the configuration settings. Defaults to True.

    Returns:
        dict: A dictionary of the custom colors used in the configuration.
    """

    RGBA = {
        k: mngs.plt.update_alpha(v, alpha)
        for k, v in mngs.plt.PARAMS["RGBA"].items()
    }
    RGBA_NORM = {
        k: tuple(mngs.plt.update_alpha(v, alpha))
        for k, v in mngs.plt.PARAMS["RGBA_NORM"].items()
    }
    RGBA_NORM_FOR_CYCLE = {
        k: tuple(mngs.plt.update_alpha(v, alpha))
        for k, v in mngs.plt.PARAMS["RGBA_NORM_FOR_CYCLE"].items()
    }  # without black, gr'a'y and white

    # Normalize figure size from mm to inches
    figsize_inch = (fig_size_mm[0] / 25.4, fig_size_mm[1] / 25.4)

    # Update Matplotlib configuration
    plt.rcParams.update(
        {
            # Resolution
            "figure.dpi": dpi_display,
            "savefig.dpi": dpi_save,
            # Figure Size
            "figure.figsize": figsize_inch,
            # # Font Size
            # "font.size": font_size_base,
            # # Title
            # "axes.titlesize": font_size_title,
            # # Axis
            # "axes.labelsize": font_size_axis_label,
            # # Ticks
            # "xtick.labelsize": font_size_tick_label,
            # "ytick.labelsize": font_size_tick_label,
            # # Legend
            # "legend.fontsize": font_size_legend,
            # Top and Right Axes
            "axes.spines.top": not hide_top_right_spines,
            "axes.spines.right": not hide_top_right_spines,
            # Custom color cycle
            "axes.prop_cycle": plt.cycler(color=RGBA_NORM_FOR_CYCLE.values()),
            # "axes.prop_cycle": plt.cycler(
            #     color=[tuple(rgba) for rgba in RGBA.values()]
            # ),
            # Line
            "lines.linewidth": line_width,
        }
    )

    if verbose:
        print("\n" + "-" * 40)
        print("Matplotlib has been configured as follows:\n")
        print(f"Figure DPI (Display): {dpi_display} DPI")
        print(f"Figure DPI (Save): {dpi_save} DPI")
        print(
            f"Figure Size (Not the Axis Size): "
            f"{fig_size_mm[0] * fig_scale:.1f} x "
            f"{fig_size_mm[1] * fig_scale:.1f} mm (width x height)"
        )
        # print(f"Font Size (Title): {font_size_title} pt")
        # print(f"Font Size (X/Y Label): {font_size_axis_label} pt")
        # print(f"Font Size (Tick Label): {font_size_tick_label} pt")
        # print(f"Font Size (Legend): {font_size_legend} pt")
        print(f"Hide Top and Right Axes: {hide_top_right_spines}")
        print(f"Custom Colors (RGBA):")
        for color_str, rgba in RGBA.items():
            print(f"  {color_str}: {rgba}")
        print("-" * 40)

    return plt, RGBA_NORM


if __name__ == "__main__":
    plt, CC = configure_mpl(plt)

    fig, axes = plt.subplots(nrows=2, sharex=True, sharey=True)
    x = np.linspace(0, 10, 100)
    for i_cc, cc_str in enumerate(CC):
        phase_shift = i_cc * np.pi / len(CC)
        y = np.sin(x + phase_shift)
        axes[0].plot(x, y, label="Default color cycle")
        axes[1].plot(x, y, color=CC[cc_str], label=f"{cc_str}")
    axes[0].legend()
    axes[1].legend()
    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-02 23:35:05 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/tex/_preview.py

import mngs
import numpy as np

def preview(tex_str_list):
    r"""
    Generate a preview of LaTeX strings.

    Example
    -------
    tex_strings = ["x^2", "\sum_{i=1}^n i"]
    fig = preview(tex_strings)
    mngs.plt.show()

    Parameters
    ----------
    tex_str_list : list of str
        List of LaTeX strings to preview

    Returns
    -------
    matplotlib.figure.Figure
        Figure containing the previews
    """
    fig, axes = mngs.plt.subplots(nrows=len(tex_str_list), ncols=1, figsize=(10, 3*len(tex_str_list)))
    axes = np.atleast_1d(axes)
    for ax, tex_string in zip(axes, tex_str_list):
        ax.text(0.5, 0.7, tex_string, size=20, ha='center', va='center')
        ax.text(0.5, 0.3, f"${tex_string}$", size=20, ha='center', va='center')
        ax.hide_spines()
    fig.tight_layout()
    return fig



#!/usr/bin/env python3

import numpy as np

from ._PARAMS import PARAMS
from matplotlib import colors as _colors


def to_rgb(c):
    return PARAMS["RGB"][c]


def to_rgba(c, alpha=0.5):
    rgba = rgb2rgba(PARAMS["RGB"][c])
    rgba[-1] = alpha
    return rgba


def to_hex(c):
    return PARAMS["HEX"][c]


def rgb2rgba(rgb, alpha=0, round=2):
    rgb = np.array(rgb).astype(float)
    rgb /= 255
    return [*rgb.round(round), alpha]


def rgba2rgb(rgba):
    rgba = np.array(rgba).astype(float)
    rgb = (rgba[:3] * 255).clip(0, 255)
    return rgb.round(2).tolist()


def update_alpha(rgba, alpha):
    rgba_list = list(rgba)
    rgba_list[-1] = alpha
    return rgba_list


def rgba2hex(rgba):
    return "#{:02x}{:02x}{:02x}{:02x}".format(
        int(rgba[0]), int(rgba[1]), int(rgba[2]), int(rgba[3] * 255)
    )


def cycle_color(i_color):
    COLORS_10_STR = list(PARAMS["RGB"].keys())
    n_colors = len(COLORS_10_STR)
    return COLORS_10_STR[i_color % n_colors]


def gradiate_color(rgb_or_rgba, n=5):
    has_alpha = len(rgb_or_rgba) == 4
    alpha = None

    if has_alpha:
        alpha = rgb_or_rgba[3]
        rgba = rgb_or_rgba
        rgb = rgba2rgb(rgba)
    else:
        rgb = rgb_or_rgba

    rgb_hsv = _colors.rgb_to_hsv([v for v in rgb])
    # gradient_dict = {}
    gradient = []

    for step in range(n):
        color_hsv = [
            rgb_hsv[0],
            rgb_hsv[1],
            rgb_hsv[2] * (1.0 - (step / n)),
        ]
        color_rgb = [int(v) for v in _colors.hsv_to_rgb(color_hsv)]
        if has_alpha:
            gradient.append(rgb2rgba(color_rgb, alpha=alpha))
        else:
            gradient.append(color_rgb)

    return gradient


if __name__ == "__main__":
    c = "blue"
    print(to_rgb(c))
    print(to_rgba(c))
    print(to_hex(c))
    print(cycle_color(1))



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-05 08:25:07 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/dsp/_detect_ripples.py


"""
This script does XYZ.
"""


"""
Imports
"""
import sys

import matplotlib.pyplot as plt
import mngs

mngs.gen.reload(mngs)

import numpy as np
import pandas as pd
from scipy.signal import find_peaks

"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def main():
    xx, tt, fs = mngs.dsp.demo_sig(sig_type="ripple")
    df = detect_ripples(xx, fs, 80, 140)
    print(df)


def detect_ripples(
    xx,
    fs,
    low_hz,
    high_hz,
    sd=2.0,
    smoothing_sigma_ms=4,
    min_duration_ms=10,
    return_preprocessed_signal=False,
):
    """
    xx: 2-dimensional (n_chs, seq_len) or 3-dimensional (batch_size, n_chs, seq_len) wide-band signal.
    """

    try:
        xx_r, fs_r = _preprocess(xx, fs, low_hz, high_hz, smoothing_sigma_ms)
        df = _find_events(xx_r, fs_r, sd, min_duration_ms)
        df = _drop_ripples_at_edges(df, low_hz, xx_r, fs_r)
        df = _calc_relative_peak_position(df)
        # df = _calc_incidence(df, xx_r, fs_r)
        df = _sort_columns(df)

        if not return_preprocessed_signal:
            return df

        elif return_preprocessed_signal:
            return df, xx_r, fs_r

    except ValueError as e:
        print("Caught an error:", e)


def _preprocess(xx, fs, low_hz, high_hz, smoothing_sigma_ms=4):
    # Ensures three dimensional
    if xx.ndim == 2:
        xx = xx[np.newaxis]
    assert xx.ndim == 3

    # For readability
    RIPPLE_BANDS = np.vstack([[low_hz, high_hz]])

    # Downsampling
    fs_tgt = low_hz * 3
    xx = mngs.dsp.resample(xx, float(fs), float(fs_tgt))
    fs = fs_tgt

    # Subtracts the global mean to reduce false detection due to EMG signal
    xx -= np.nanmean(xx, axis=1, keepdims=True)

    # Bandpass Filtering
    xx = (
        (
            mngs.dsp.filt.bandpass(
                np.array(xx),
                fs_tgt,
                RIPPLE_BANDS,
            )
        )
        .squeeze(-2)
        .astype(np.float64)
    )

    # Calculate RMS
    xx = xx**2
    _, xx = mngs.dsp.hilbert(xx)
    xx = mngs.dsp.filt.gauss(xx, smoothing_sigma_ms * 1e-3 * fs_tgt).squeeze(
        -2
    )
    xx = np.sqrt(xx)

    # Scales across channels
    xx = xx.mean(axis=1)
    xx = mngs.gen.to_z(xx, dim=-1)

    return xx, fs_tgt


def _find_events(xx_r, fs_r, sd, min_duration_ms):
    def _find_events_1d(xx_ri, fs_r, sd, min_duration_ms):
        # Finds peaks over the designated standard deviation
        peaks, properties = find_peaks(xx_ri, height=sd)

        # Determines the range around each peak (customize as needed)
        peaks_all = []
        peak_ranges = []
        peak_amplitudes_sd = []

        for peak in peaks:
            left_bound = np.where(xx_ri[:peak] < 0)[0]
            right_bound = np.where(xx_ri[peak:] < 0)[0]

            left_ips = left_bound.max() if left_bound.size > 0 else peak
            right_ips = (
                peak + right_bound.min() if right_bound.size > 0 else peak
            )

            # Avoid duplicates: Check if the current peak range is already listed
            if not any(
                (left_ips == start and right_ips == end)
                for start, end in peak_ranges
            ):
                peaks_all.append(peak)
                peak_ranges.append((left_ips, right_ips))
                peak_amplitudes_sd.append(xx_ri[peak])

        # Converts to DataFrame
        if peak_ranges:
            starts, ends = zip(*peak_ranges) if peak_ranges else ([], [])
            df = pd.DataFrame(
                {
                    "start_s": np.hstack(starts) / fs_r,
                    "peak_s": np.hstack(peaks_all) / fs_r,
                    "end_s": np.hstack(ends) / fs_r,
                    "peak_amp_sd": np.hstack(peak_amplitudes_sd),
                }
            ).round(3)
        else:
            df = pd.DataFrame(
                columns=["start_s", "peak_s", "end_s", "peak_amp_sd"]
            )

        # Duration
        df["duration_s"] = df.end_s - df.start_s

        # Filters events with short duration
        df = df[df.duration_s > (min_duration_ms * 1e-3)]

        return df

    if xx_r.ndim == 1:
        xx_r = xx_r[np.newaxis, :]
    assert xx_r.ndim == 2

    dfs = []
    for i_ch in range(len(xx_r)):
        xx_ri = xx_r[i_ch]
        df_i = _find_events_1d(xx_ri, fs_r, sd, min_duration_ms)
        df_i.index = [i_ch for _ in range(len(df_i))]
        dfs.append(df_i)
    dfs = pd.concat(dfs)

    return dfs


def _drop_ripples_at_edges(df, low_hz, xx_r, fs_r):
    edge_s = 1 / low_hz * 3
    indi_drop = (df.start_s < edge_s) + (
        xx_r.shape[-1] / fs_r - edge_s < df.end_s
    )
    df = df[~indi_drop]
    return df


def _calc_relative_peak_position(df):
    delta_s = df.peak_s - df.start_s
    rel_peak = delta_s / df.duration_s
    df["rel_peak_pos"] = np.round(rel_peak, 3)
    return df


# def _calc_incidence(df, xx_r, fs_r):
#     n_ripples = len(df)
#     rec_s = xx_r.shape[-1] / fs_r
#     df["incidence_hz"] = n_ripples / rec_s
#     return df


def _sort_columns(df):
    sorted_columns = [
        "start_s",
        "end_s",
        "duration_s",
        "peak_s",
        "rel_peak_pos",
        "peak_amp_sd",
        # "incidence_hz",
    ]
    df = df[sorted_columns]
    return df


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF


# # append iEEG traces
# iEEG_traces = []
# for (
#     i_rip,
#     rip,
# ) in (
#     rip_df.reset_index().iterrows()
# ):  # rip_df = rips_df.iloc[i_trial]

#     start_pts = int(rip["start_time"] * CONFIG["FS_iEEG"])
#     end_pts = int(rip["end_time"] * CONFIG["FS_iEEG"])
#     # start_pts = int(rip["start_time"] * FS_TGT)
#     # end_pts = int(rip["end_time"] * FS_TGT)
#     iEEG_traces.append(
#         # iEEG[i_trial][:, start_pts:end_pts]
#         np.array(iEEG)[i_trial][:, start_pts:end_pts]
#     )
# rip_df["iEEG trace"] = iEEG_traces


# # append ripple band filtered iEEG traces
# ripple_band_iEEG_traces = []
# for (
#     i_rip,
#     rip,
# ) in (
#     rip_df.reset_index().iterrows()
# ):  # rip_df = rips_df.iloc[i_trial]
#     start_pts = int(rip["start_time"] * CONFIG["FS_iEEG"])
#     end_pts = int(rip["end_time"] * CONFIG["FS_iEEG"])

#     ripple_band_iEEG_traces.append(
#         iEEG_ripple_band_passed[i_trial][:, start_pts:end_pts]
#     )
# rip_df["ripple band iEEG trace"] = ripple_band_iEEG_traces

# # ripple peak amplitude
# ripple_peak_amplitude = [
#     np.abs(rbt).max(axis=-1) for rbt in ripple_band_iEEG_traces
# ]
# ripple_band_baseline_sd = iEEG_ripple_band_passed[i_trial].std(
#     axis=-1
# )
# rip_df["ripple_peak_amplitude_sd"] = [
#     (rpa / ripple_band_baseline_sd).mean()
#     for rpa in ripple_peak_amplitude
# ]

# rip_df["ripple_amplitude_sd"] = [
#     (np.abs(rbt).mean(axis=-1) / ripple_band_baseline_sd).mean()
#     for rbt in rip_df["ripple band iEEG trace"]
# ]



#!/usr/bin/env python3

import json
import os
import pickle
import warnings
from glob import glob

import h5py
import joblib
import mne
import mngs
import numpy as np
import pandas as pd
import torch
import yaml
from docx import Document
from openpyxl import load_workbook
from PIL import Image


def load(lpath, show=False, verbose=False, **kwargs):
    """
    Load data from various file formats.

    This function supports loading data from multiple file formats including CSV, Excel, Numpy, Pickle, JSON, YAML, and more.

    Parameters:
    -----------
    lpath : str
        The path to the file to be loaded.
    show : bool, optional
        If True, display additional information during loading. Default is False.
    verbose : bool, optional
        If True, print verbose output during loading. Default is False.
    **kwargs : dict
        Additional keyword arguments to be passed to the specific loading function.

    Returns:
    --------
    object
        The loaded data object, which can be of various types depending on the input file format.

    Raises:
    -------
    ValueError
        If the file extension is not supported.

    Examples:
    ---------
    >>> data = load('data.csv')
    >>> image = load('image.png')
    >>> model = load('model.pth')
    """
    if lpath.startswith('f"'):
        try:
            lpath = eval(lpath)
        except:
            lpath = lpath.replace('f"', '')

    lpath = lpath.replace("/./", "/")

    try:
        extension = "." + lpath.split(".")[-1]

        # CSV
        if extension == ".csv":
            index_col = kwargs.get("index_col", 0)
            obj = pd.read_csv(lpath, **kwargs)
            obj = obj.loc[:, ~obj.columns.str.contains("^Unnamed")]
        # TSV
        elif extension == ".tsv":
            obj = pd.read_csv(lpath, sep="\t", **kwargs)
        # Excel
        elif extension in [".xls", ".xlsx", ".xlsm", ".xlsb"]:
            obj = pd.read_excel(lpath, **kwargs)
        # Parquet
        elif extension == ".parquet":
            obj = pd.read_parquet(lpath, **kwargs)
        # Numpy
        elif extension == ".npy":
            obj = np.load(lpath, allow_pickle=True, **kwargs)
        # Numpy NPZ
        elif extension == ".npz":
            obj = np.load(lpath)
            obj = dict(obj)
            obj = [v for v in obj.values()]
        # Pickle
        elif extension == ".pkl":
            with open(lpath, "rb") as l:
                obj = pickle.load(l, **kwargs)
        # Joblib
        elif extension == ".joblib":
            with open(lpath, "rb") as l:
                obj = joblib.load(l, **kwargs)
        # HDF5
        elif extension == ".hdf5":
            obj = {}
            with h5py.File(lpath, "r") as hf:
                for name in hf:
                    obj[name] = hf[name][:]
        # JSON
        elif extension == ".json":
            with open(lpath, "r") as f:
                obj = json.load(f)
        # Image
        elif extension in [".jpg", ".png", ".tiff", "tif"]:
            obj = Image.open(lpath)
        # YAML
        elif extension in [".yaml", ".yml"]:
            lower = kwargs.pop("lower", False)
            with open(lpath) as f:
                obj = yaml.safe_load(f, **kwargs)
            if lower:
                obj = {k.lower(): v for k, v in obj.items()}
        # Text
        elif extension in [".txt", ".log", ".event"]:
            with open(lpath, "r") as f:
                obj = f.read().splitlines()
                # obj = tqdm(f.read().splitlines(), desc=f"Reading {lpath}")
        # Markdown
        elif extension == ".md":
            obj = load_markdown(lpath, **kwargs)
        # PyTorch
        elif extension in [".pth", ".pt"]:
            obj = torch.load(lpath, **kwargs)
        # MATLAB
        elif extension == ".mat":
            from pymatreader import read_mat

            obj = read_mat(lpath, **kwargs)
        # XML
        elif extension == ".xml":
            from xml2dict import xml2dict

            obj = xml2dict(lpath, **kwargs)
        # CON
        elif extension == ".con":
            obj = mne.io.read_raw_fif(lpath, preload=True, **kwargs)
            obj = obj.to_data_frame()
            obj["samp_rate"] = obj.info["sfreq"]
        # CatBoost model
        elif extension == ".cbm":
            from catboost import CatBoostModel

            obj = CatBoostModel.load_model(lpath, **kwargs)
        # EEG data
        elif extension in [
            ".vhdr",
            ".vmrk",
            ".edf",
            ".bdf",
            ".gdf",
            ".cnt",
            ".egi",
            ".eeg",
            ".set",
        ]:
            obj = _load_eeg_data(lpath, **kwargs)
        else:
            raise ValueError(f"Unsupported file extension: {extension}")

        return obj
    except Exception as e:
        print(f"Error loading file {lpath}: {str(e)}")
        raise


def _load_text(lpath):
    """
    Load text from a file.

    Parameters:
    -----------
    lpath : str
        The path to the text file to be loaded.

    Returns:
    --------
    str
        The content of the text file as a string.

    Raises:
    -------
    FileNotFoundError
        If the specified file does not exist.
    IOError
        If there's an error reading the file.
    """
    with open(lpath, "r") as f:
        return f.read()


def _load_eeg_data(filename, **kwargs):
    """
    Load EEG data based on file extension and associated files using MNE-Python.

    This function supports various EEG file formats including BrainVision, EDF, BDF, GDF, CNT, EGI, and SET.
    It also handles special cases for .eeg files (BrainVision and Nihon Koden).

    Parameters:
    -----------
    filename : str
        The path to the EEG file to be loaded.
    **kwargs : dict
        Additional keyword arguments to be passed to the specific MNE loading function.

    Returns:
    --------
    raw : mne.io.Raw
        The loaded raw EEG data.

    Raises:
    -------
    ValueError
        If the file extension is not supported.

    Notes:
    ------
    This function uses MNE-Python to load the EEG data. It automatically detects the file format
    based on the file extension and uses the appropriate MNE function to load the data.
    """
    # Get the file extension
    extension = filename.split(".")[-1]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)

        # Load the data based on the file extension
        if extension in ["vhdr", "vmrk"]:
            # Load BrainVision data
            raw = mne.io.read_raw_brainvision(filename, preload=True, **kwargs)
        elif extension == "edf":
            # Load European data format
            raw = mne.io.read_raw_edf(filename, preload=True, **kwargs)
        elif extension == "bdf":
            # Load BioSemi data format
            raw = mne.io.read_raw_bdf(filename, preload=True, **kwargs)
        elif extension == "gdf":
            # Load General data format
            raw = mne.io.read_raw_gdf(filename, preload=True, **kwargs)
        elif extension == "cnt":
            # Load Neuroscan CNT data
            raw = mne.io.read_raw_cnt(filename, preload=True, **kwargs)
        elif extension == "egi":
            # Load EGI simple binary data
            raw = mne.io.read_raw_egi(filename, preload=True, **kwargs)
        elif extension == "set":
            # ???
            raw = mne.io.read_raw(filename, preload=True, **kwargs)
        elif extension == "eeg":
            is_BrainVision = any(
                os.path.isfile(filename.replace(".eeg", ext))
                for ext in [".vhdr", ".vmrk"]
            )
            is_NihonKoden = any(
                os.path.isfile(filename.replace(".eeg", ext))
                for ext in [".21e", ".pnt", ".log"]
            )

            # Brain Vision
            if is_BrainVision:
                filename_v = filename.replace(".eeg", ".vhdr")
                raw = mne.io.read_raw_brainvision(
                    filename_v, preload=True, **kwargs
                )
            # Nihon Koden
            if is_NihonKoden:
                # raw = mne.io.read_raw_nihon(filename, preload=True, **kwargs)
                raw = mne.io.read_raw(filename, preload=True, **kwargs)
        else:
            raise ValueError(f"Unsupported file extension: {extension}")

        return raw


def load_markdown(lpath_md, style="plain_text"):
    """
    Load and convert Markdown content from a file.

    This function reads a Markdown file and converts it to either HTML or plain text format.

    Parameters:
    -----------
    lpath_md : str
        The path to the Markdown file to be loaded.
    style : str, optional
        The output style of the converted content.
        Options are "html" or "plain_text" (default).

    Returns:
    --------
    str
        The converted content of the Markdown file, either as HTML or plain text.

    Raises:
    -------
    FileNotFoundError
        If the specified file does not exist.
    IOError
        If there's an error reading the file.
    ValueError
        If an invalid style option is provided.

    Notes:
    ------
    This function uses the 'markdown' library to convert Markdown to HTML,
    and 'html2text' to convert HTML to plain text when necessary.
    """
    import html2text
    import markdown

    # Load Markdown content from a file
    with open(lpath_md, "r") as file:
        markdown_content = file.read()

    # Convert Markdown to HTML
    html_content = markdown.markdown(markdown_content)
    if style == "html":
        return html_content
    elif style == "plain_text":
        text_maker = html2text.HTML2Text()
        text_maker.ignore_links = True
        text_maker.bypass_tables = False
        plain_text = text_maker.handle(html_content)
        return plain_text
    else:
        raise ValueError(
            "Invalid style option. Choose 'html' or 'plain_text'."
        )


def _check_encoding(file_path):
    """
    Check the encoding of a given file.

    This function attempts to read the file with different encodings
    to determine the correct one.

    Parameters:
    -----------
    file_path : str
        The path to the file to check.

    Returns:
    --------
    str
        The detected encoding of the file.

    Raises:
    -------
    IOError
        If the file cannot be read or the encoding cannot be determined.
    """
    import chardet

    with open(file_path, "rb") as file:
        raw_data = file.read()

    result = chardet.detect(raw_data)
    return result["encoding"]


#     from chardet.universaldetector import UniversalDetector

#     detector = UniversalDetector()
#     with open(file_path, mode="rb") as f:
#         for binary in f:
#             detector.feed(binary)
#             if detector.done:
#                 break
#     detector.close()
#     enc = detector.result["encoding"]
#     return enc


def get_data_path_from_a_package(package_str, resource):
    """
    Get the path to a data file within a package.

    This function finds the path to a data file within a package's data directory.

    Parameters:
    -----------
    package_str : str
        The name of the package as a string.
    resource : str
        The name of the resource file within the package's data directory.

    Returns:
    --------
    str
        The full path to the resource file.

    Raises:
    -------
    ImportError
        If the specified package cannot be found.
    FileNotFoundError
        If the resource file does not exist in the package's data directory.
    """
    import importlib
    import os
    import sys

    spec = importlib.util.find_spec(package_str)
    if spec is None:
        raise ImportError(f"Package '{package_str}' not found")

    data_dir = os.path.join(spec.origin.split("src")[0], "data")
    resource_path = os.path.join(data_dir, resource)

    if not os.path.exists(resource_path):
        raise FileNotFoundError(
            f"Resource '{resource}' not found in package '{package_str}'"
        )

    return resource_path


def load_yaml_as_an_optuna_dict(fpath_yaml, trial):
    """
    Load a YAML file and convert it to an Optuna-compatible dictionary.

    This function reads a YAML file containing hyperparameter configurations
    and converts it to a dictionary suitable for use with Optuna trials.

    Parameters:
    -----------
    fpath_yaml : str
        The file path to the YAML configuration file.
    trial : optuna.trial.Trial
        The Optuna trial object to use for suggesting hyperparameters.

    Returns:
    --------
    dict
        A dictionary containing the hyperparameters with values suggested by Optuna.

    Raises:
    -------
    FileNotFoundError
        If the specified YAML file does not exist.
    ValueError
        If the YAML file contains invalid configuration for Optuna.
    """
    _d = load(fpath_yaml)

    for k, v in _d.items():
        dist = v["distribution"]

        if dist == "categorical":
            _d[k] = trial.suggest_categorical(k, v["values"])

        elif dist == "uniform":
            _d[k] = trial.suggest_int(k, float(v["min"]), float(v["max"]))

        elif dist == "loguniform":
            _d[k] = trial.suggest_loguniform(
                k, float(v["min"]), float(v["max"])
            )

        elif dist == "intloguniform":
            _d[k] = trial.suggest_int(
                k, float(v["min"]), float(v["max"]), log=True
            )

    return _d


def load_study_rdb(study_name, rdb_raw_bytes_url):
    """
    Load an Optuna study from a RDB (Relational Database) file.

    This function loads an Optuna study from a given RDB file URL.

    Parameters:
    -----------
    study_name : str
        The name of the Optuna study to load.
    rdb_raw_bytes_url : str
        The URL of the RDB file, typically in the format "sqlite:///*.db".

    Returns:
    --------
    optuna.study.Study
        The loaded Optuna study object.

    Raises:
    -------
    optuna.exceptions.StorageInvalidUsageError
        If there's an error loading the study from the storage.

    Example:
    --------
    >>> study = load_study_rdb(
    ...     study_name="YOUR_STUDY_NAME",
    ...     rdb_raw_bytes_url="sqlite:///path/to/your/study.db"
    ... )
    """
    import optuna

    storage = optuna.storages.RDBStorage(url=rdb_raw_bytes_url)
    study = optuna.load_study(study_name=study_name, storage=storage)
    print(f"Loaded: {rdb_raw_bytes_url}")
    return study


# def load_configs(IS_DEBUG=None, show=False, verbose=False):
#     """
#     Load and process configuration files from the ./config directory.

#     This function loads all YAML files in the ./config directory, processes debug configurations,
#     and returns a consolidated configuration dictionary.

#     Parameters:
#     -----------
#     IS_DEBUG : bool, optional
#         Flag to enable debug mode. If None, it will be determined from the ./config/IS_DEBUG.yaml file.
#     show : bool, optional
#         If True, print debug information. Default is False.
#     verbose : bool, optional
#         If True, print verbose debug information. Default is False.

#     Returns:
#     --------
#     mngs.gen.DotDict
#         A dictionary-like object containing all loaded and processed configurations.

#     Notes:
#     ------
#     - If the CI environment variable is set to "True", IS_DEBUG will be set to True.
#     - Debug configurations (keys starting with "DEBUG_") will replace their non-debug counterparts when IS_DEBUG is True.
#     """

#     if os.getenv("CI") == "True":
#         IS_DEBUG = True

#     def use_debug_values(config, IS_DEBUG):
#         """
#         Update configuration with debug values if IS_DEBUG is True.

#         Parameters:
#         -----------
#         config : dict
#             The configuration dictionary to update.
#         IS_DEBUG : bool
#             Flag indicating whether debug mode is active.

#         Returns:
#         --------
#         dict
#             The updated configuration dictionary.
#         """
#         if IS_DEBUG:
#             debug_keys = mngs.gen.search("^DEBUG_", list(config.keys()))[1]
#             for dk in debug_keys:
#                 dk_wo_debug_prefix = dk.split("DEBUG_")[1]
#                 config[dk_wo_debug_prefix] = config[dk]
#                 if show or verbose:
#                     print(f"\n{dk} -> {dk_wo_debug_prefix}\n")
#         return config

#     # Check ./config/IS_DEBUG.yaml file if IS_DEBUG argument is not passed
#     if IS_DEBUG is None:
#         IS_DEBUG_PATH = "./config/IS_DEBUG.yaml"
#         if os.path.exists(IS_DEBUG_PATH):
#             IS_DEBUG = mngs.io.load("./config/IS_DEBUG.yaml").get("IS_DEBUG")
#         else:
#             IS_DEBUG = False

#     # Main
#     CONFIGS = {}
#     for lpath in glob("./config/*.yaml"):
#         CONFIG = use_debug_values(mngs.io.load(lpath), IS_DEBUG)
#         CONFIGS.update(CONFIG)

#     CONFIGS = mngs.gen.DotDict(CONFIGS)

#     return CONFIGS


################################################################################
# dev
################################################################################
def _load_docx(lpath):
    """
    Load and extract text content from a .docx file.

    Parameters:
    -----------
    lpath : str
        The path to the .docx file.

    Returns:
    --------
    str
        The extracted text content from the .docx file.

    Raises:
    -------
    FileNotFoundError
        If the specified file does not exist.
    docx.opc.exceptions.PackageNotFoundError
        If the file is not a valid .docx file.
    """
    from docx import Document

    doc = Document(lpath)
    full_text = []
    for para in doc.paragraphs:
        full_text.append(para.text)
    return "".join(full_text)


#     doc = docx.Document(lpath)
#     full_text = []
#     for para in doc.paragraphs:
#         full_text.append(para.text)
#     return "\n".join(full_text)


# def _load_pdf(lpath):
#     reader = PyPDF2.PdfReader(lpath)
#     full_text = []
#     for page_num in range(len(reader.pages)):
#         page = reader.pages[page_num]
#         full_text.append(page.extract_text())
#     return "\n".join(full_text)


# def _load_latex(lpath):
#     return lpath.read().decode("utf-8")


# def _load_excel(lpath):
#     workbook = openpyxl.load_workbook(lpath)
#     all_text = []
#     for sheet in workbook:
#         for row in sheet.iter_rows(values_only=True):
#             all_text.append(
#                 " ".join(
#                     [str(cell) if cell is not None else "" for cell in row]
#                 )
#             )
#     return "\n".join(all_text)


# def _load_markdown(lpath):
#     md_text = StringIO(lpath.read().decode("utf-8"))
#     html = markdown.markdown(md_text.read())
#     return html


# def _load_textfile(lpath):
#     return lpath.read().decode("utf-8")


def load_markdown(lpath_md, style="plain_text"):
    """
    Load and convert a Markdown file to either HTML or plain text.

    Parameters:
    -----------
    lpath_md : str
        The path to the Markdown file.
    style : str, optional
        The output style, either "html" or "plain_text" (default).

    Returns:
    --------
    str
        The converted content of the Markdown file.
    """
    import html2text
    import markdown

    # Load Markdown content from a file
    with open(lpath_md, "r") as file:
        markdown_content = file.read()

    # Convert Markdown to HTML
    html_content = markdown.markdown(markdown_content)
    if style == "html":
        return html_content

    elif style == "plain_text":
        text_maker = html2text.HTML2Text()
        text_maker.ignore_links = True
        text_maker.bypass_tables = False
        plain_text = text_maker.handle(html_content)

        return plain_text


def load_yaml_as_an_optuna_dict(fpath_yaml, trial):
    """
    Load a YAML file and convert it to an Optuna-compatible dictionary.

    Parameters:
    -----------
    fpath_yaml : str
        The path to the YAML file.
    trial : optuna.trial.Trial
        The Optuna trial object.

    Returns:
    --------
    dict
        A dictionary with Optuna-compatible parameter suggestions.
    """
    _d = load(fpath_yaml)

    for k, v in _d.items():

        dist = v["distribution"]

        if dist == "categorical":
            _d[k] = trial.suggest_categorical(k, v["values"])

        elif dist == "uniform":
            _d[k] = trial.suggest_int(k, float(v["min"]), float(v["max"]))

        elif dist == "loguniform":
            _d[k] = trial.suggest_loguniform(
                k, float(v["min"]), float(v["max"])
            )

        elif dist == "intloguniform":
            _d[k] = trial.suggest_int(
                k, float(v["min"]), float(v["max"]), log=True
            )

    return _d


def load_study_rdb(study_name, rdb_raw_bytes_url):
    """
    Load an Optuna study from a RDB storage.

    Parameters:
    -----------
    study_name : str
        The name of the Optuna study.
    rdb_raw_bytes_url : str
        The URL of the RDB storage.

    Returns:
    --------
    optuna.study.Study
        The loaded Optuna study object.
    """
    import optuna

    # rdb_raw_bytes_url = "sqlite:////tmp/fake/ywatanabe/_MicroNN_WindowSize-1.0-sec_MaxEpochs_100_2021-1216-1844/optuna_study_test_file#0.db"
    storage = optuna.storages.RDBStorage(url=rdb_raw_bytes_url)
    study = optuna.load_study(study_name=study_name, storage=storage)
    print(f"\nLoaded: {rdb_raw_bytes_url}\n")
    return study


def load_configs(IS_DEBUG=None, show=False, verbose=False):
    """
    Load configuration files from the ./config directory.

    Parameters:
    -----------
    IS_DEBUG : bool, optional
        If True, use debug configurations. If None, check ./config/IS_DEBUG.yaml.
    show : bool, optional
        If True, display additional information during loading.
    verbose : bool, optional
        If True, print verbose output during loading.

    Returns:
    --------
    mngs.gen.DotDict
        A dictionary-like object containing the loaded configurations.
    """

    def apply_debug_values(config, IS_DEBUG):
        if IS_DEBUG:
            if isinstance(config, (dict, mngs.gen.DotDict)):
                for key, value in list(config.items()):
                    try:
                        if key.startswith("DEBUG_"):
                            dk_wo_debug_prefix = key.split("DEBUG_")[1]
                            config[dk_wo_debug_prefix] = value
                            if show or verbose:
                                print(f"\n{key} -> {dk_wo_debug_prefix}\n")
                        elif isinstance(value, (dict, mngs.gen.DotDict)):
                            config[key] = apply_debug_values(value, IS_DEBUG)
                    except Exception as e:
                        print(e)
        return config

    if os.getenv("CI") == "True":
        IS_DEBUG = True

    try:
        # Check ./config/IS_DEBUG.yaml file if IS_DEBUG argument is not passed
        if IS_DEBUG is None:
            IS_DEBUG_PATH = "./config/IS_DEBUG.yaml"
            if os.path.exists(IS_DEBUG_PATH):
                IS_DEBUG = mngs.io.load("./config/IS_DEBUG.yaml").get("IS_DEBUG")
            else:
                IS_DEBUG = False

        # Main
        CONFIGS = {}
        for lpath in glob("./config/*.yaml"):
            config = mngs.io.load(lpath)
            if config:
                CONFIG = apply_debug_values(config, IS_DEBUG)
                CONFIGS.update(CONFIG)

        CONFIGS = mngs.gen.DotDict(CONFIGS)

    except Exception as e:
        print(e)
        CONFIGS = mngs.gen.DotDict({})

    return CONFIGS



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-05-30 09:37:02 (ywatanabe)"


import mngs
import numpy as np
from mngs.general import torch_fn
from mngs.nn import (
    BandPassFilter,
    BandStopFilter,
    GaussianFilter,
    HighPassFilter,
    LowPassFilter,
)


@torch_fn
def gauss(x, sigma, t=None):
    return GaussianFilter(sigma)(x, t=t)


@torch_fn
def bandpass(x, fs, bands, t=None):
    return BandPassFilter(bands, fs, x.shape[-1])(x, t=t)


@torch_fn
def bandstop(x, fs, bands, t=None):
    return BandStopFilter(bands, fs, x.shape[-1])(x, t=t)


@torch_fn
def lowpass(x, fs, cutoffs_hz, t=None):
    return LowPassFilter(cutoffs_hz, fs, x.shape[-1])(x, t=t)


@torch_fn
def highpass(x, fs, cutoffs_hz, t=None):
    return HighPassFilter(cutoffs_hz, fs, x.shape[-1])(x, t=t)


def _custom_print(x):
    print(type(x), x.shape)


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt
    import mngs
    import torch

    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # Parametes
    T_SEC = 1
    SRC_FS = 1024
    FREQS_HZ = list(np.linspace(0, 500, 10, endpoint=False).astype(int))
    SIG_TYPE = "periodic"
    BANDS = np.vstack([[80, 310]])
    SIGMA = 3

    # Demo Signal
    xx, tt, fs = mngs.dsp.demo_sig(
        t_sec=T_SEC,
        fs=SRC_FS,
        freqs_hz=FREQS_HZ,
        sig_type=SIG_TYPE,
    )

    # Filtering
    x_bp, t_bp = mngs.dsp.filt.bandpass(xx, fs, BANDS, t=tt)
    x_bs, t_bs = mngs.dsp.filt.bandstop(xx, fs, BANDS, t=tt)
    x_lp, t_lp = mngs.dsp.filt.lowpass(xx, fs, BANDS[:, 0], t=tt)
    x_hp, t_hp = mngs.dsp.filt.highpass(xx, fs, BANDS[:, 1], t=tt)
    x_g, t_g = mngs.dsp.filt.gauss(xx, sigma=SIGMA, t=tt)
    filted = {
        f"Original (Sum of {FREQS_HZ}-Hz signals)": (xx, tt, fs),
        f"Bandpass-filtered ({BANDS[0][0]} - {BANDS[0][1]} Hz)": (
            x_bp,
            t_bp,
            fs,
        ),
        f"Bandstop-filtered ({BANDS[0][0]} - {BANDS[0][1]} Hz)": (
            x_bs,
            t_bs,
            fs,
        ),
        f"Lowpass-filtered ({BANDS[0][0]} Hz)": (x_lp, t_lp, fs),
        f"Highpass-filtered ({BANDS[0][1]} Hz)": (x_hp, t_hp, fs),
        f"Gaussian-filtered (sigma = {SIGMA} SD [point])": (x_g, t_g, fs),
    }

    # Plots traces
    fig, axes = plt.subplots(
        nrows=len(filted), ncols=1, sharex=True, sharey=True
    )
    i_batch = 0
    i_ch = 0
    i_filt = 0
    for ax, (k, v) in zip(axes, filted.items()):
        _xx, _tt, _fs = v
        if _xx.ndim == 3:
            _xx = _xx[i_batch, i_ch]
        elif _xx.ndim == 4:
            _xx = _xx[i_batch, i_ch, i_filt]
        ax.plot(_tt, _xx, label=k)
        ax.legend(loc="upper left")

    fig.suptitle("Filtered")
    fig.supxlabel("Time [s]")
    fig.supylabel("Amplitude")

    mngs.io.save(fig, "traces.png")

    # Calculates and Plots PSD
    fig, axes = plt.subplots(
        nrows=len(filted), ncols=1, sharex=True, sharey=True
    )
    i_batch = 0
    i_ch = 0
    i_filt = 0
    for ax, (k, v) in zip(axes, filted.items()):
        _xx, _tt, _fs = v

        _psd, ff = mngs.dsp.psd(_xx, _fs)
        if _psd.ndim == 3:
            _psd = _psd[i_batch, i_ch]
        elif _psd.ndim == 4:
            _psd = _psd[i_batch, i_ch, i_filt]

        ax.plot(ff, _psd, label=k)
        ax.legend(loc="upper left")

        for bb in np.hstack(BANDS):
            ax.axvline(x=bb, color=CC["grey"], linestyle="--")

    fig.suptitle("PSD (power spectrum density) of filtered signals")
    fig.supxlabel("Frequency [Hz]")
    fig.supylabel("log(Power [uV^2 / Hz]) [a.u.]")
    mngs.io.save(fig, "psd.png")

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/home/ywatanabe/proj/mngs/src/mngs/dsp/filt.py
"""
